import platform
import struct

import mpitestutil as testutil
import mpiunittest as unittest

from mpi4py import MPI

datatypes_c = [
    MPI.CHAR,
    MPI.WCHAR,
    MPI.SIGNED_CHAR,
    MPI.SHORT,
    MPI.INT,
    MPI.LONG,
    MPI.UNSIGNED_CHAR,
    MPI.UNSIGNED_SHORT,
    MPI.UNSIGNED,
    MPI.UNSIGNED_LONG,
    MPI.LONG_LONG,
    MPI.UNSIGNED_LONG_LONG,
    MPI.FLOAT,
    MPI.DOUBLE,
    MPI.LONG_DOUBLE,
]
datatypes_c99 = [
    MPI.C_BOOL,
    MPI.INT8_T,
    MPI.INT16_T,
    MPI.INT32_T,
    MPI.INT64_T,
    MPI.UINT8_T,
    MPI.UINT16_T,
    MPI.UINT32_T,
    MPI.UINT64_T,
    MPI.FLOAT16_T,
    MPI.FLOAT32_T,
    MPI.FLOAT64_T,
    MPI.BFLOAT16_T,
    MPI.C_COMPLEX,
    MPI.C_FLOAT_COMPLEX,
    MPI.C_DOUBLE_COMPLEX,
    MPI.C_LONG_DOUBLE_COMPLEX,
]
datatypes_f77 = [
    MPI.CHARACTER,
    MPI.LOGICAL,
    MPI.INTEGER,
    MPI.REAL,
    MPI.DOUBLE_PRECISION,
    MPI.COMPLEX,
    MPI.DOUBLE_COMPLEX,
]
datatypes_f90 = [
    MPI.LOGICAL1,
    MPI.LOGICAL2,
    MPI.LOGICAL4,
    MPI.LOGICAL8,
    MPI.LOGICAL16,
    MPI.INTEGER1,
    MPI.INTEGER2,
    MPI.INTEGER4,
    MPI.INTEGER8,
    MPI.INTEGER16,
    MPI.REAL2,
    MPI.REAL4,
    MPI.REAL8,
    MPI.REAL16,
    MPI.COMPLEX4,
    MPI.COMPLEX8,
    MPI.COMPLEX16,
    MPI.COMPLEX32,
]
datatypes_mpi = [
    MPI.PACKED,
    MPI.BYTE,
    MPI.AINT,
    MPI.OFFSET,
]

for typelist in [
    datatypes_c,
    datatypes_c99,
    datatypes_f77,
    datatypes_f90,
]:
    typelist[:] = [t for t in typelist if testutil.has_datatype(t)]
del typelist

datatypes = []
datatypes += datatypes_c
datatypes += datatypes_c99
datatypes += datatypes_f77
datatypes += datatypes_f90
datatypes += datatypes_mpi

combiner_map = {}

machine = platform.machine()


class TestDatatypeNull(unittest.TestCase):
    #
    def testConstructor(self):
        datatype = MPI.Datatype()
        self.assertEqual(datatype, MPI.DATATYPE_NULL)
        self.assertIsNot(datatype, MPI.DATATYPE_NULL)

        def construct():
            MPI.Datatype((1, 2, 3))

        self.assertRaises(TypeError, construct)

    def testGetName(self):
        name = MPI.DATATYPE_NULL.Get_name()
        self.assertEqual(name, "MPI_DATATYPE_NULL")


class TestDatatype(unittest.TestCase):
    #
    def testBoolEqNe(self):
        for dtype in datatypes:
            self.assertTrue(not not dtype)
            eq = dtype == MPI.Datatype(dtype)
            ne = dtype != MPI.Datatype(dtype)
            self.assertTrue(eq)
            self.assertFalse(ne)

    def testGetExtent(self):
        for dtype in datatypes:
            lb, ext = dtype.Get_extent()
            self.assertEqual(dtype.lb, lb)
            self.assertEqual(dtype.ub, lb + ext)
            self.assertEqual(dtype.extent, ext)

    def testGetSize(self):
        for dtype in datatypes:
            size = dtype.Get_size()
            self.assertEqual(dtype.size, size)

    def testGetTrueExtent(self):
        for dtype in datatypes:
            try:
                lb, ext = dtype.Get_true_extent()
                self.assertEqual(dtype.true_lb, lb)
                self.assertEqual(dtype.true_ub, lb + ext)
                self.assertEqual(dtype.true_extent, ext)
            except NotImplementedError:
                self.skipTest("mpi-type-get_true_extent")

    match_size_logical = [1, 2, 4, 8, 16]
    match_size_integer = [1, 2, 4, 8, 16]
    match_size_real = [2, 4, 8]
    match_size_complex = [4, 8, 16]

    @unittest.skipMPI("MPI(<2.0)")
    @unittest.skipMPI("openmpi", not testutil.has_datatype(MPI.INTEGER))
    def testMatchSize(self):
        for key in ("logical", "integer", "real", "complex"):
            match_size = getattr(self, f"match_size_{key}")
            typeclass = getattr(MPI, f"TYPECLASS_{key.upper()}")
            if typeclass == MPI.UNDEFINED:
                continue
            for size in match_size:
                try:
                    datatype = MPI.Datatype.Match_size(typeclass, size)
                except MPI.Exception:
                    datatype = getattr(MPI, f"{key.upper()}{size}")
                    if testutil.has_datatype(datatype):
                        raise
                else:
                    if testutil.has_datatype(datatype):
                        self.assertEqual(datatype.size, size)

    def testGetValueIndex(self):
        typenames = ("SHORT", "INT", "LONG", "FLOAT", "DOUBLE", "LONG_DOUBLE")
        value_types = [getattr(MPI, f"{attr}") for attr in typenames]
        pair_types = [getattr(MPI, f"{attr}_INT") for attr in typenames]
        for value, pair in zip(value_types, pair_types):
            result = MPI.Datatype.Get_value_index(value, MPI.INT)
            self.assertEqual(result, pair)
        for value in value_types:
            result = MPI.Datatype.Get_value_index(value, MPI.FLOAT)
            self.assertEqual(result, MPI.DATATYPE_NULL)
        with self.assertRaises(MPI.Exception) as catcher:
            MPI.Datatype.Get_value_index(MPI.DATATYPE_NULL, MPI.INT)
        self.assertEqual(catcher.exception.Get_error_class(), MPI.ERR_TYPE)
        with self.assertRaises(MPI.Exception) as catcher:
            MPI.Datatype.Get_value_index(MPI.INT, MPI.DATATYPE_NULL)
        self.assertEqual(catcher.exception.Get_error_class(), MPI.ERR_TYPE)

    def testGetEnvelope(self):
        for dtype in datatypes:
            try:
                envelope = dtype.Get_envelope()
            except NotImplementedError:
                self.skipTest("mpi-type-get_envelope")
            ni, na, nc, nd, combiner = envelope
            self.assertEqual(combiner, MPI.COMBINER_NAMED)
            self.assertEqual(ni, 0)
            self.assertEqual(na, 0)
            self.assertEqual(nc, 0)
            self.assertEqual(nd, 0)
            self.assertEqual(dtype.envelope, envelope)
            self.assertEqual(dtype.combiner, combiner)
            self.assertTrue(dtype.is_named)
            self.assertTrue(dtype.is_predefined)
            otype, combiner, params = dtype.decode()
            self.assertIs(dtype, otype)
            self.assertEqual(combiner, "NAMED")
            self.assertEqual(params, {})

    def testGetSetName(self):
        name = MPI.DATATYPE_NULL.Get_name()
        self.assertEqual(name, "MPI_DATATYPE_NULL")
        for dtype in datatypes:
            try:
                name = dtype.Get_name()
                self.assertTrue(name)
                dtype.Set_name(name)
                self.assertEqual(name, dtype.Get_name())
                dtype.name = dtype.name
            except NotImplementedError:
                self.skipTest("mpi-type-name")

    def testCommit(self):
        for dtype in datatypes:
            dtype.Commit()

    @unittest.skipMPI("mpich(<4.2.3)", machine.startswith("ppc"))
    @unittest.skipMPI("openmpi(<6.0.0)", machine in {"ppc", "ppc64"})
    def testCodeCharStr(self):
        f90datatypes = []
        try:
            try:
                for r in (1, 2, 4):
                    f90datatypes.append(MPI.Datatype.Create_f90_integer(r))
                for p, r in ((6, 30), (15, 300)):
                    f90datatypes.append(MPI.Datatype.Create_f90_real(p, r))
                for p, r in ((6, 30), (15, 300)):
                    f90datatypes.append(MPI.Datatype.Create_f90_complex(p, r))
            except MPI.Exception:
                if not unittest.is_mpi("msmpi"):
                    raise
            f90datatypes = [
                dtype for dtype in f90datatypes if dtype and dtype.size > 0
            ]
        except NotImplementedError:
            f90datatypes = []
        largef90datatypes = []
        if MPI.INTEGER16 != MPI.DATATYPE_NULL:
            largef90datatypes += [MPI.INTEGER16]
        if (
            struct.calcsize("P") == 4
            or MPI.DOUBLE.extent == MPI.LONG_DOUBLE.extent
        ):
            largef90datatypes += [MPI.REAL16, MPI.COMPLEX32]
        for dtype in datatypes + f90datatypes:
            with self.subTest(datatype=dtype.name or "f90"):
                if dtype == MPI.BFLOAT16_T:
                    continue
                if dtype in largef90datatypes:
                    continue
                code = dtype.tocode()
                self.assertIsNotNone(code)
                mpitype = MPI.Datatype.fromcode(code)
                self.assertEqual(dtype.typechar, mpitype.typechar)
                self.assertEqual(dtype.typestr, mpitype.typestr)
                try:
                    mpitypedup1 = mpitype.Dup()
                    self.assertEqual(mpitypedup1.tocode(), mpitype.tocode())
                    self.assertEqual(mpitypedup1.typestr, mpitype.typestr)
                    self.assertEqual(mpitypedup1.typechar, mpitype.typechar)
                    mpitypedup2 = mpitypedup1.Dup()
                    self.assertEqual(mpitypedup2.tocode(), mpitype.tocode())
                    self.assertEqual(mpitypedup2.typestr, mpitype.typestr)
                    self.assertEqual(mpitypedup2.typechar, mpitype.typechar)
                finally:
                    mpitypedup1.Free()
                    mpitypedup2.Free()
        with self.assertRaises(ValueError):
            MPI.Datatype.fromcode("abc@xyz")
        with self.assertRaises(ValueError):
            MPI.DATATYPE_NULL.tocode()
        with self.assertRaises(ValueError):
            MPI.INT_INT.tocode()
        self.assertEqual(MPI.INT_INT.typechar, "V")
        self.assertEqual(MPI.INT_INT.typestr, f"V{MPI.INT.extent * 2}")

    @unittest.skipMPI("mpich", struct.calcsize("P") == 4)
    def testContiguousBigMPI(self):
        int_max = (1 << (struct.calcsize("i") * 8 - 1)) - 1
        mpitype = MPI.BYTE.Create_contiguous(int_max)
        self.assertEqual(mpitype.size, int_max)
        basetype, combiner, params = mpitype.decode()
        self.assertEqual(basetype, MPI.BYTE)
        self.assertEqual(combiner, "CONTIGUOUS")
        self.assertEqual(params, {"count": int_max})
        mpitype.Free()
        if MPI.Get_version() < (3, 1):
            return
        count = int_max + int_max // 1024
        mpitype = MPI.BYTE.Create_contiguous(count)
        self.assertEqual(mpitype.size, count)
        basetype, combiner, params = mpitype.decode()
        mpitype.Free()
        if MPI.Get_version() >= (4, 0) or unittest.is_mpi("impi(>2021.12.0)"):
            self.assertEqual(basetype, MPI.BYTE)
            self.assertEqual(combiner, "CONTIGUOUS")
            self.assertEqual(params, {"count": count})
        else:
            self.assertEqual(basetype, MPI.DATATYPE_NULL)
            self.assertEqual(combiner, "STRUCT")
            qtype, rtype = params["datatypes"]
            self.assertEqual(qtype.combiner, MPI.COMBINER_VECTOR)
            self.assertEqual(qtype.size, (count // int_max) * int_max)
            self.assertEqual(rtype.combiner, MPI.COMBINER_CONTIGUOUS)
            self.assertEqual(rtype.size, count % int_max)
            qtype.free()
            rtype.free()


class BaseTestDatatypeCreateMixin:
    #
    def free(self, newtype):
        if newtype == MPI.DATATYPE_NULL:
            return
        *_, combiner = newtype.Get_envelope()
        if combiner in (
            MPI.COMBINER_NAMED,
            MPI.COMBINER_F90_INTEGER,
            MPI.COMBINER_F90_REAL,
            MPI.COMBINER_F90_COMPLEX,
        ):
            return
        newtype.Free()

    def check_contents(self, _factory, newtype, _oldtype):
        try:
            envelope = newtype.Get_envelope()
            contents = newtype.Get_contents()
        except NotImplementedError:
            self.skipTest("mpi-type-get_envelope")
        ni, na, nc, nd, combiner = envelope
        i, a, c, d = contents
        self.assertEqual(ni, len(i))
        self.assertEqual(na, len(a))
        self.assertEqual(nc, len(c))
        self.assertEqual(nd, len(d))
        self.assertNotEqual(combiner, MPI.COMBINER_NAMED)
        self.assertEqual(newtype.envelope, envelope)
        self.assertEqual(newtype.combiner, combiner)
        self.assertFalse(newtype.is_named)
        if combiner in (
            MPI.COMBINER_F90_INTEGER,
            MPI.COMBINER_F90_REAL,
            MPI.COMBINER_F90_COMPLEX,
        ):
            self.assertTrue(newtype.is_predefined)
        else:
            self.assertFalse(newtype.is_predefined)
        for dt in d:
            self.free(dt)
        contents = newtype.contents
        self.assertEqual(contents[:-1], (i, a, c))
        for dt in contents[-1]:
            self.free(dt)

    def check_recreate(self, factory, newtype):
        name = factory.__name__
        name = name.replace("Get_value_index", "Create_value_index")
        NAME = name.replace("Create_", "").upper()
        symbol = getattr(MPI, "COMBINER_" + NAME)
        if symbol == MPI.UNDEFINED:
            return
        if combiner_map is None:
            return
        symbol = combiner_map.get(symbol, symbol)
        if symbol is None:
            return
        self.assertEqual(symbol, newtype.combiner)
        decoded1 = newtype.decode()
        oldtype, constructor, kwargs = decoded1
        prefix = "create" if constructor != "VALUE_INDEX" else "get"
        constructor = prefix.title() + "_" + constructor.lower()
        newtype2 = getattr(oldtype, constructor)(**kwargs)
        decoded2 = newtype2.decode()
        types1 = decoded1[2].pop("datatypes", [])
        types2 = decoded2[2].pop("datatypes", [])
        for dt1, dt2 in zip(types1, types2):
            self.assertEqual(dt1.combiner, dt2.combiner)
            self.assertEqual(dt1.typechar, dt2.typechar)
            self.assertEqual(dt1.typestr, dt2.typestr)
            self.free(dt1)
            self.free(dt2)
        self.assertEqual(decoded1[1], decoded2[1])
        self.assertEqual(decoded2[2], decoded2[2])
        for dec in (decoded1, decoded2):
            self.free(dec[0])
        self.free(newtype2)

    def testDup(self):
        for dtype in datatypes:
            factory = MPI.Datatype.Dup
            self.check(dtype, factory)

    def testContiguous(self):
        for dtype in datatypes:
            for count in range(5):
                factory = MPI.Datatype.Create_contiguous
                args = (count,)
                self.check(dtype, factory, *args)

    def testVector(self):
        for dtype in datatypes:
            for count in range(5):
                for blocklength in range(5):
                    for stride in range(5):
                        factory = MPI.Datatype.Create_vector
                        args = (count, blocklength, stride)
                        self.check(dtype, factory, *args)

    def testHvector(self):
        for dtype in datatypes:
            for count in range(5):
                for blocklength in range(5):
                    for stride in range(5):
                        factory = MPI.Datatype.Create_hvector
                        args = (count, blocklength, stride)
                        self.check(dtype, factory, *args)

    def testIndexed(self):
        for dtype in datatypes:
            for block in range(5):
                blocklengths = list(range(block, block + 5))
                displacements = [0]
                for b in blocklengths[:-1]:
                    stride = displacements[-1] + b * dtype.extent + 1
                    displacements.append(stride)
                factory = MPI.Datatype.Create_indexed
                args = (blocklengths, displacements)
                self.check(dtype, factory, *args)
                # args = (block, displacements) XXX
                # self.check(dtype, factory, *args)  XXX

    def testIndexedBlock(self):
        for dtype in datatypes:
            for block in range(5):
                blocklengths = list(range(block, block + 5))
                displacements = [0]
                for b in blocklengths[:-1]:
                    stride = displacements[-1] + b * dtype.extent + 1
                    displacements.append(stride)
                factory = MPI.Datatype.Create_indexed_block
                args = (block, displacements)
                self.check(dtype, factory, *args)

    def testHindexed(self):
        for dtype in datatypes:
            for block in range(5):
                blocklengths = list(range(block, block + 5))
                displacements = [0]
                for b in blocklengths[:-1]:
                    stride = displacements[-1] + b * dtype.extent + 1
                    displacements.append(stride)

                factory = MPI.Datatype.Create_hindexed
                args = (blocklengths, displacements)
                self.check(dtype, factory, *args)
                # args = (block, displacements) XXX
                # self.check(dtype, factory, *args)  XXX

    @unittest.skipMPI("openmpi(<=1.8.1)", MPI.Get_version()[0] == 3)
    def testHindexedBlock(self):
        for dtype in datatypes:
            for block in range(5):
                displacements = [0]
                for _i in range(5):
                    stride = displacements[-1] + block * dtype.extent + 1
                    displacements.append(stride)
                factory = MPI.Datatype.Create_hindexed_block
                args = (block, displacements)
                self.check(dtype, factory, *args)

    def testStruct(self):
        for dtype1 in datatypes:
            for dtype2 in datatypes:
                dtypes = (dtype1, dtype2)
                blocklengths = (2, 3)
                displacements = [0]
                for dtype in dtypes[:-1]:
                    stride = displacements[-1] + dtype.extent
                    displacements.append(stride)
                factory = MPI.Datatype.Create_struct
                args = (blocklengths, displacements, dtypes)
                self.check(None, factory, *args)
        for dtype in datatypes:
            factory = MPI.Datatype.Create_struct
            dtypes = [dtype.Dup()]
            dtypes.append(dtypes[-1].Create_contiguous(2))
            dtypes.append(dtypes[-1].Dup())
            dtypes.append(dtypes[-1].Create_struct([1], [0], [dtypes[-1]]))
            dtypes.append(dtypes[-1].Dup())
            dtypes.append(dtypes[-1].Create_resized(0, dtypes[-1].extent))
            dtypes.append(dtypes[-1].Dup())
            for dt in dtypes:
                args = [[1, 1], [0, dt.extent * 2], (dt, dt)]
                self.check(None, factory, *args)
                dt.Free()
        with self.assertRaises(ValueError):
            factory = MPI.Datatype.Create_struct
            factory([1], [0], [MPI.INT, MPI.FLOAT])

    def testSubarray(self):
        for dtype in datatypes:
            for ndim in range(1, 5):
                for size in range(1, 5):
                    for subsize in range(1, size):
                        for start in range(size - subsize):
                            for order in [
                                MPI.ORDER_C,
                                MPI.ORDER_FORTRAN,
                                MPI.ORDER_F,
                            ]:
                                sizes = [size] * ndim
                                subsizes = [subsize] * ndim
                                starts = [start] * ndim
                                factory = MPI.Datatype.Create_subarray
                                args = sizes, subsizes, starts, order
                                self.check(dtype, factory, *args)

    def testDarray(self):
        for dtype in datatypes:
            for ndim in range(1, 3 + 1):
                for size in (4, 8, 9, 27):
                    for rank in (0, size - 1):
                        for dist in [
                            MPI.DISTRIBUTE_BLOCK,
                            MPI.DISTRIBUTE_CYCLIC,
                        ]:
                            for order in [MPI.ORDER_C, MPI.ORDER_F]:
                                gsizes = [size] * ndim
                                distribs = [dist] * ndim
                                dargs = [MPI.DISTRIBUTE_DFLT_DARG] * ndim
                                psizes = MPI.Compute_dims(size, [0] * ndim)
                                factory = MPI.Datatype.Create_darray
                                args = (
                                    size,
                                    rank,
                                    gsizes,
                                    distribs,
                                    dargs,
                                    psizes,
                                    order,
                                )
                                self.check(dtype, factory, *args)

    def testF90Integer(self):
        for r in (1, 2, 4):
            factory = MPI.Datatype.Create_f90_integer
            args = (r,)
            self.check(None, factory, *args)

    @unittest.skipMPI("openmpi(<3.0.0)")
    @unittest.skipMPI("msmpi")
    def testF90RealSingle(self):
        (p, r) = (6, 30)
        factory = MPI.Datatype.Create_f90_real
        args = (p, r)
        self.check(None, factory, *args)

    @unittest.skipMPI("mpich(<4.2.3)", machine.startswith("ppc"))
    @unittest.skipMPI("openmpi(<6.0.0)", machine in {"ppc", "ppc64"})
    @unittest.skipMPI("openmpi(<3.0.0)")
    @unittest.skipMPI("msmpi")
    def testF90RealDouble(self):
        (p, r) = (15, 300)
        factory = MPI.Datatype.Create_f90_real
        args = (p, r)
        self.check(None, factory, *args)

    @unittest.skipMPI("openmpi(<3.0.0)")
    @unittest.skipMPI("msmpi")
    def testF90ComplexSingle(self):
        (p, r) = (6, 30)
        factory = MPI.Datatype.Create_f90_complex
        args = (p, r)
        self.check(None, factory, *args)

    @unittest.skipMPI("mpich(<4.2.3)", machine.startswith("ppc"))
    @unittest.skipMPI("openmpi(<6.0.0)", machine in {"ppc", "ppc64"})
    @unittest.skipMPI("openmpi(<3.0.0)")
    @unittest.skipMPI("msmpi")
    def testF90ComplexDouble(self):
        (p, r) = (15, 300)
        factory = MPI.Datatype.Create_f90_complex
        args = (p, r)
        self.check(None, factory, *args)

    def testResized(self):
        for dtype in datatypes:
            for lb in range(-10, 10):
                for extent in range(1, 10):
                    factory = MPI.Datatype.Create_resized
                    args = lb, extent
                    self.check(dtype, factory, *args)

    def testValueIndex(self):
        integral_types = datatypes_c[2:-3] + datatypes_c99[1:9]
        floating_types = datatypes_c[-3:]
        value_types = integral_types + floating_types
        index_types = integral_types
        for value in value_types:
            if value == MPI.DATATYPE_NULL:
                continue
            for index in index_types:
                if index == MPI.DATATYPE_NULL:
                    continue
                factory = MPI.Datatype.Get_value_index
                pair = factory(value, index)
                if pair == MPI.DATATYPE_NULL:
                    continue
                if pair.is_named:
                    continue
                self.check(None, factory, value, index)


class TestDatatypeCreate(BaseTestDatatypeCreateMixin, unittest.TestCase):
    #
    def check(self, oldtype, factory, *args):
        try:
            if oldtype is not None:
                newtype = factory(oldtype, *args)
            else:
                newtype = factory(*args)
            if newtype == MPI.DATATYPE_NULL:
                return
        except NotImplementedError:
            self.skipTest("mpi-type-constructor")
        self.check_contents(factory, newtype, oldtype)
        self.check_recreate(factory, newtype)
        newtype.Commit()
        self.check_contents(factory, newtype, oldtype)
        self.check_recreate(factory, newtype)
        self.free(newtype)


class TestDatatypePickle(BaseTestDatatypeCreateMixin, unittest.TestCase):
    #
    def check(self, oldtype, factory, *args):
        from pickle import dumps, loads

        try:
            if oldtype is not None:
                newtype0 = factory(oldtype, *args)
            else:
                newtype0 = factory(*args)
            if newtype0 == MPI.DATATYPE_NULL:
                return
        except NotImplementedError:
            self.skipTest("mpi-type-constructor")
        newtype1 = loads(dumps(newtype0))
        self.check_contents(factory, newtype1, oldtype)
        self.free(newtype1)
        self.free(newtype0)

    def testNamed(self):
        from pickle import dumps, loads

        for dtype in [MPI.DATATYPE_NULL, *datatypes]:
            newdtype = loads(dumps(dtype))
            self.assertIs(newdtype, dtype)
            newdtype = loads(dumps(MPI.Datatype(dtype)))
            self.assertIsNot(newdtype, dtype)
            self.assertEqual(newdtype, dtype)


name, version = MPI.get_vendor()
if name == "MPICH1":
    combiner_map[MPI.COMBINER_VECTOR] = None
    combiner_map[MPI.COMBINER_HVECTOR] = None
    combiner_map[MPI.COMBINER_INDEXED] = None
    combiner_map[MPI.COMBINER_HINDEXED_BLOCK] = None
    for t in datatypes_f77:
        if t in datatypes:
            datatypes.remove(t)
    datatypes_f77.clear()
elif MPI.Get_version() < (2, 0):
    combiner_map = None
if name == "Open MPI":
    if (1, 6, 0) < version < (1, 7, 0):
        TestDatatype.match_size_complex[:] = []
    if version < (1, 5, 2):
        for t in [getattr(MPI, f"COMPLEX{i}") for i in (4, 8, 16, 32)]:
            if t in datatypes:
                datatypes.remove(t)
            if t in datatypes_f90:
                datatypes_f90.remove(t)
if name in ("MPICH", "Intel MPI"):
    if MPI.FLOAT16_T != MPI.DATATYPE_NULL:
        if MPI.FLOAT16_T.Get_name() == "":
            MPI.FLOAT16_T.Set_name("MPIX_C_FLOAT16")


if __name__ == "__main__":
    unittest.main()



import optparse
import pathlib
import sys

from mpi4py import MPI

assert __name__ == "__main__"
assert sys.path[0] == str(pathlib.Path(__file__).parent)
if pathlib.Path(__file__).name == "__main__.py":
    assert sys.argv[0] == sys.path[0]
else:
    assert sys.argv[0] == __file__

parser = optparse.OptionParser()
parser.add_option(
    "--rank",
    action="store",
    type="int",
    dest="rank",
    default=0,
)
parser.add_option(
    "--interrupt",
    action="store_true",
    dest="interrupt",
    default=False,
)
parser.add_option(
    "--sys-exit",
    action="store",
    type="int",
    dest="sys_exit",
    default=None,
)
parser.add_option(
    "--sys-exit-msg",
    action="store",
    type="string",
    dest="sys_exit",
    default=None,
)
parser.add_option(
    "--exception",
    action="store",
    type="string",
    dest="exception",
    default=None,
)
(options, args) = parser.parse_args()
assert not args

comm = MPI.COMM_WORLD

comm.Barrier()
comm.Barrier()
if comm.rank == options.rank:
    if options.interrupt:
        raise KeyboardInterrupt
    if options.sys_exit:
        sys.exit(options.sys_exit)
    if options.exception:
        raise RuntimeError(options.exception)

comm.Barrier()
comm.Barrier()
if comm.rank > 0:
    comm.Recv([None, "B"], comm.rank - 1)
print("Hello, World!", flush=True)
if comm.rank < comm.size - 1:
    comm.Send([None, "B"], comm.rank + 1)
comm.Barrier()

sys.exit()



cdef class Request:
    """
    Request handler.
    """

    def __cinit__(self, Request request: Request | None = None):
        cinit(self, request)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Request): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_default(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Request:
        """
        Create object from MPI handle.
        """
        if issubclass(cls, Prequest):
            return PyMPIPrequest_New(<MPI_Request> <Py_uintptr_t> handle)
        if issubclass(cls, Grequest):
            return PyMPIGrequest_New(<MPI_Request> <Py_uintptr_t> handle)
        return fromhandle(<MPI_Request> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Free` if not null.
        """
        safefree(self)

    # Completion Operations
    # ---------------------

    def Wait(self, Status status: Status | None = None) -> Literal[True]:
        """
        Wait for a non-blocking operation to complete.
        """
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Wait(
            &self.ob_mpi, statusp) )
        if self.ob_mpi == MPI_REQUEST_NULL:
            self.ob_buf = None
        return True

    def Test(self, Status status: Status | None = None) -> bool:
        """
        Test for the completion of a non-blocking operation.
        """
        cdef int flag = 0
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Test(
            &self.ob_mpi, &flag, statusp) )
        if self.ob_mpi == MPI_REQUEST_NULL:
            self.ob_buf = None
        return <bint>flag

    def Get_status(self, Status status: Status | None = None) -> bool:
        """
        Non-destructive test for the completion of a request.
        """
        cdef int flag = 0
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Request_get_status(
            self.ob_mpi, &flag, statusp) )
        return <bint>flag

    # Multiple Completions
    # --------------------

    @classmethod
    def Waitany(
        cls,
        requests: Sequence[Request],
        Status status: Status | None = None,
    ) -> int:
        """
        Wait for any previously initiated request to complete.
        """
        cdef int index = MPI_UNDEFINED
        cdef MPI_Status *statusp = arg_Status(status)
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests)
        try:
            with nogil: CHKERR( MPI_Waitany(
                rs.count, rs.requests, &index, statusp) )
        finally:
            rs.release()
        return index

    @classmethod
    def Testany(
        cls,
        requests: Sequence[Request],
        Status status: Status | None = None,
    ) -> tuple[int, bool]:
        """
        Test for completion of any previously initiated request.
        """
        cdef int index = MPI_UNDEFINED
        cdef int flag = 0
        cdef MPI_Status *statusp = arg_Status(status)
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests)
        try:
            with nogil: CHKERR( MPI_Testany(
                rs.count, rs.requests, &index, &flag, statusp) )
        finally:
            rs.release()
        return (index, <bint>flag)

    @classmethod
    def Get_status_any(
        cls,
        requests: Sequence[Request],
        Status status: Status | None = None,
    ) -> tuple[int, bool]:
        """
        Non-destructive test for the completion of any requests.
        """
        cdef int index = MPI_UNDEFINED
        cdef int flag = 0
        cdef MPI_Status *statusp = arg_Status(status)
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests)
        try:
            with nogil: CHKERR( MPI_Request_get_status_any(
                rs.count, rs.requests, &index, &flag, statusp) )
        finally:
            rs.release()            # ~> MPI-4.1
        return (index, <bint>flag)  # ~> MPI-4.1

    @classmethod
    def Waitall(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> Literal[True]:
        """
        Wait for all previously initiated requests to complete.
        """
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests, statuses)
        try:
            with nogil: CHKERR( MPI_Waitall(
                rs.count, rs.requests, rs.statuses) )
        finally:
            rs.release(statuses)
        return True

    @classmethod
    def Testall(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> bool:
        """
        Test for completion of all previously initiated requests.
        """
        cdef int flag = 0
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests, statuses)
        try:
            with nogil: CHKERR( MPI_Testall(
                rs.count, rs.requests, &flag, rs.statuses) )
        finally:
            rs.release(statuses)
        return <bint>flag

    @classmethod
    def Get_status_all(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> bool:
        """
        Non-destructive test for the completion of all requests.
        """
        cdef int flag = 0
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests, statuses)
        try:
            with nogil: CHKERR( MPI_Request_get_status_all(
                rs.count, rs.requests, &flag, rs.statuses) )
        finally:
            rs.release(statuses)  # ~> MPI-4.1
        return <bint>flag         # ~> MPI-4.1

    @classmethod
    def Waitsome(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> list[int] | None:
        """
        Wait for some previously initiated requests to complete.
        """
        cdef object indices = None
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests, statuses)
        rs.add_indices()
        try:
            with nogil: CHKERR( MPI_Waitsome(
                rs.count, rs.requests, &rs.outcount, rs.indices, rs.statuses) )
            indices = rs.get_indices()
        finally:
            rs.release(statuses)
        return indices

    @classmethod
    def Testsome(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> list[int] | None:
        """
        Test for completion of some previously initiated requests.
        """
        cdef object indices = None
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests, statuses)
        rs.add_indices()
        try:
            with nogil: CHKERR( MPI_Testsome(
                rs.count, rs.requests, &rs.outcount, rs.indices, rs.statuses) )
            indices = rs.get_indices()
        finally:
            rs.release(statuses)
        return indices

    @classmethod
    def Get_status_some(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> list[int] | None:
        """
        Non-destructive test for completion of some requests.
        """
        cdef object indices = None
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests, statuses)
        rs.add_indices()
        try:
            with nogil: CHKERR( MPI_Request_get_status_some(
                rs.count, rs.requests, &rs.outcount, rs.indices, rs.statuses) )
            indices = rs.get_indices()  # ~> MPI-4.1
        finally:
            rs.release(statuses)  # ~> MPI-4.1
        return indices            # ~> MPI-4.1

    # Cancel
    # ------

    def Cancel(self) -> None:
        """
        Cancel a request.
        """
        with nogil: CHKERR( MPI_Cancel(&self.ob_mpi) )

    # Deallocation
    # ------------

    def Free(self) -> None:
        """
        Free a communication request.
        """
        cdef MPI_Request save = self.ob_mpi
        with nogil: CHKERR( MPI_Request_free(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Request_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Request:
        """
        Translate integer handle to object.
        """
        if issubclass(cls, Prequest):
            return PyMPIPrequest_New(MPI_Request_fromint(arg))
        if issubclass(cls, Grequest):
            return PyMPIGrequest_New(MPI_Request_fromint(arg))
        return fromhandle(MPI_Request_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Request_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Request:
        """
        """
        if issubclass(cls, Prequest):
            return PyMPIPrequest_New(MPI_Request_f2c(arg))
        if issubclass(cls, Grequest):
            return PyMPIGrequest_New(MPI_Request_f2c(arg))
        return fromhandle(MPI_Request_f2c(arg))

    # Python Communication
    # --------------------

    def wait(
        self,
        Status status: Status | None = None,
    ) -> Any:
        """
        Wait for a non-blocking operation to complete.
        """
        cdef msg = PyMPI_wait(self, status)
        return msg

    def test(
        self,
            Status status: Status | None = None,
    ) -> tuple[bool, Any | None]:
        """
        Test for the completion of a non-blocking operation.
        """
        cdef int flag = 0
        cdef msg = PyMPI_test(self, &flag, status)
        return (<bint>flag, msg)

    def get_status(
        self,
        Status status: Status | None = None,
    ) -> bool:
        """
        Non-destructive test for the completion of a request.
        """
        cdef int flag = 0
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Request_get_status(
            self.ob_mpi, &flag, statusp) )
        return <bint>flag

    @classmethod
    def waitany(
        cls,
        requests: Sequence[Request],
        Status status: Status | None = None
    ) -> tuple[int, Any]:
        """
        Wait for any previously initiated request to complete.
        """
        cdef int index = MPI_UNDEFINED
        cdef msg = PyMPI_waitany(requests, &index, status)
        return (index, msg)

    @classmethod
    def testany(
        cls,
        requests: Sequence[Request],
        Status status: Status | None = None,
    ) -> tuple[int, bool, Any | None]:
        """
        Test for completion of any previously initiated request.
        """
        cdef int index = MPI_UNDEFINED
        cdef int flag  = 0
        cdef msg = PyMPI_testany(requests, &index, &flag, status)
        return (index, <bint>flag, msg)

    @classmethod
    def get_status_any(
        cls,
        requests: Sequence[Request],
        Status status: Status | None = None,
    ) -> tuple[int, bool]:
        """
        Non-destructive test for the completion of any requests.
        """
        return Request.Get_status_any(requests, status)

    @classmethod
    def waitall(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> list[Any]:
        """
        Wait for all previously initiated requests to complete.
        """
        cdef msg = PyMPI_waitall(requests, statuses)
        return msg

    @classmethod
    def testall(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None
    ) -> tuple[bool, list[Any] | None]:
        """
        Test for completion of all previously initiated requests.
        """
        cdef int flag = 0
        cdef msg = PyMPI_testall(requests, &flag, statuses)
        return (<bint>flag, msg)

    @classmethod
    def get_status_all(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> bool:
        """
        Non-destructive test for the completion of all requests.
        """
        return Request.Get_status_all(requests, statuses)

    @classmethod
    def waitsome(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> tuple[list[int] | None, list[Any] | None]:
        """
        Wait for some previously initiated requests to complete.
        """
        return PyMPI_waitsome(requests, statuses)

    @classmethod
    def testsome(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> tuple[list[int] | None, list[Any] | None]:
        """
        Test for completion of some previously initiated requests.
        """
        return PyMPI_testsome(requests, statuses)

    @classmethod
    def get_status_some(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> list[int] | None:
        """
        Non-destructive test for completion of some requests.
        """
        return Request.Get_status_some(requests, statuses)

    def cancel(self) -> None:
        """
        Cancel a request.
        """
        with nogil: CHKERR( MPI_Cancel(&self.ob_mpi) )


cdef class Prequest(Request):
    """
    Persistent request handler.
    """

    def __cinit__(self, Request request: Request | None = None):
        if self.ob_mpi == MPI_REQUEST_NULL: return
        <void>(<Prequest?>request)

    def Start(self) -> None:
        """
        Initiate a communication with a persistent request.
        """
        with nogil: CHKERR( MPI_Start(&self.ob_mpi) )

    @classmethod
    def Startall(cls, requests: list[Prequest]) -> None:
        """
        Start a collection of persistent requests.
        """
        cdef _p_rs rs = _p_rs.__new__(_p_rs)
        rs.acquire(requests)
        try:
            with nogil: CHKERR( MPI_Startall(rs.count, rs.requests) )
        finally:
            rs.release()

    # Partitioned completion
    # ----------------------

    def Pready(
        self,
        int partition: int,
    ) -> None:
        """
        Mark a given partition as ready.
        """
        CHKERR( MPI_Pready(partition, self.ob_mpi) )

    def Pready_range(
        self,
        int partition_low: int,
        int partition_high: int,
    ) -> None:
        """
        Mark a range of partitions as ready.
        """
        CHKERR( MPI_Pready_range(partition_low, partition_high, self.ob_mpi) )

    def Pready_list(
        self,
        partitions: Sequence[int],
    ) -> None:
        """
        Mark a sequence of partitions as ready.
        """
        cdef int length = 0, *array_of_partitions = NULL
        partitions = getarray(partitions, &length, &array_of_partitions)
        CHKERR( MPI_Pready_list(length, array_of_partitions, self.ob_mpi) )

    def Parrived(
        self,
        int partition: int,
    ) -> bool:
        """
        Test partial completion of a partitioned receive operation.
        """
        cdef int flag = 0
        CHKERR( MPI_Parrived(self.ob_mpi, partition, &flag) )
        return <bint>flag


cdef class Grequest(Request):
    """
    Generalized request handler.
    """

    def __cinit__(self, Request request: Request | None = None):
        self.ob_grequest = self.ob_mpi
        if self.ob_mpi == MPI_REQUEST_NULL: return
        <void>(<Grequest?>request)

    @classmethod
    def Start(
        cls,
        query_fn: Callable[..., None] | None = None,
        free_fn: Callable[..., None] | None = None,
        cancel_fn: Callable[..., None] | None = None,
        args: tuple[Any] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> Grequest:
        """
        Create and return a user-defined request.
        """
        cdef Grequest request = <Grequest>New(Grequest)
        cdef _p_greq state = _p_greq(
            query_fn, free_fn, cancel_fn, args, kwargs)
        with nogil: CHKERR( MPI_Grequest_start(
            greq_query_fn, greq_free_fn, greq_cancel_fn,
            <void*>state, &request.ob_mpi) )
        Py_INCREF(state)
        request.ob_grequest = request.ob_mpi
        return request

    def Complete(self) -> None:
        """
        Notify that a user-defined request is complete.
        """
        if self.ob_mpi != MPI_REQUEST_NULL:
            if self.ob_mpi != self.ob_grequest:
                raise MPIException(MPI_ERR_REQUEST)  # ~> unreachable
        cdef MPI_Request grequest = self.ob_grequest
        self.ob_grequest = self.ob_mpi  # or MPI_REQUEST_NULL ??
        with nogil: CHKERR( MPI_Grequest_complete(grequest) )
        self.ob_grequest = self.ob_mpi  # or MPI_REQUEST_NULL ??

    def complete(self, obj: Any = None) -> None:
        """
        Notify that a user-defined request is complete.
        """
        self.ob_buf = PyMPI_wrap_object(obj)
        Grequest.Complete(self)


cdef Request __REQUEST_NULL__ = def_Request( MPI_REQUEST_NULL , "REQUEST_NULL" )


# Predefined request handles
# --------------------------

REQUEST_NULL = __REQUEST_NULL__  #: Null request handle



import mpiunittest as unittest

from mpi4py import MPI

_basic = [
    None,
    True,
    False,
    -7,
    0,
    7,
    -(2**63) + 1,
    2**63 - 1,
    -2.17,
    0.0,
    +1.62,
    1 + 2j,
    2 - 3j,
    "mpi4py",
]
messages = list(_basic)
messages += [
    list(_basic),
    tuple(_basic),
    {f"k{k}": v for k, v in enumerate(_basic)},
]


@unittest.skipIf(MPI.MESSAGE_NULL == MPI.MESSAGE_NO_PROC, "mpi-message")
class TestMessage(unittest.TestCase):
    #
    def testMessageNull(self):
        null = MPI.MESSAGE_NULL
        self.assertFalse(null)
        null2 = MPI.Message()
        self.assertEqual(null, null2)
        null3 = MPI.Message(null)
        self.assertEqual(null, null3)

    def testMessageNoProc(self):
        #
        noproc = MPI.MESSAGE_NO_PROC
        self.assertTrue(noproc)
        noproc.recv()
        self.assertTrue(noproc)
        noproc.irecv().wait()
        self.assertTrue(noproc)
        #
        noproc2 = MPI.Message(MPI.MESSAGE_NO_PROC)
        self.assertTrue(noproc2)
        self.assertEqual(noproc2, noproc)
        self.assertNotEqual(noproc, MPI.MESSAGE_NULL)
        #
        message = MPI.Message(MPI.MESSAGE_NO_PROC)
        message.recv()
        self.assertEqual(message, MPI.MESSAGE_NULL)
        #
        message = MPI.Message(MPI.MESSAGE_NO_PROC)
        request = message.irecv()
        self.assertEqual(message, MPI.MESSAGE_NULL)
        self.assertNotEqual(request, MPI.REQUEST_NULL)
        request.wait()
        self.assertEqual(request, MPI.REQUEST_NULL)
        #
        comm = MPI.COMM_SELF
        message = comm.mprobe(MPI.PROC_NULL)
        self.assertNotEqual(message, MPI.MESSAGE_NULL)
        self.assertEqual(message, MPI.MESSAGE_NO_PROC)
        noproc = comm.improbe(MPI.PROC_NULL)
        self.assertNotEqual(message, MPI.MESSAGE_NULL)
        self.assertEqual(message, MPI.MESSAGE_NO_PROC)


@unittest.skipIf(MPI.MESSAGE_NULL == MPI.MESSAGE_NO_PROC, "mpi-message")
class BaseTestP2PMatched:
    #
    COMM = MPI.COMM_NULL

    def testIMProbe(self):
        comm = self.COMM.Dup()
        try:
            m = comm.improbe()
            self.assertIsNone(m)
            m = comm.improbe(MPI.ANY_SOURCE)
            self.assertIsNone(m)
            m = comm.improbe(MPI.ANY_SOURCE, MPI.ANY_TAG)
            self.assertIsNone(m)
            status = MPI.Status()
            m = comm.improbe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertIsNone(m)
            self.assertEqual(status.source, MPI.ANY_SOURCE)
            self.assertEqual(status.tag, MPI.ANY_TAG)
            self.assertEqual(status.error, MPI.SUCCESS)
            m = MPI.Message.iprobe(comm)
            self.assertIsNone(m)
            s = comm.isend(None, comm.rank, 0)
            r = comm.mprobe(comm.rank, 0).irecv()
            MPI.Request.waitall([s, r])
        finally:
            comm.Free()

    def testProbeRecv(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        for smsg in messages:
            if size == 1:
                sr = comm.isend(smsg, 0, 0)
                m = comm.mprobe(0, 0)
                self.assertIsInstance(m, MPI.Message)
                self.assertTrue(m)
                rr = m.irecv()
                self.assertFalse(m)
                self.assertTrue(sr)
                self.assertTrue(rr)
                MPI.Request.Waitall([sr, rr])
                self.assertFalse(sr)
                self.assertFalse(rr)
                #
                r = comm.isend(smsg, 0, 0)
                m = MPI.Message.probe(comm, 0, 0)
                self.assertIsInstance(m, MPI.Message)
                self.assertTrue(m)
                rmsg = m.recv()
                self.assertFalse(m)
                r.wait()
            elif rank == 0:
                comm.send(smsg, 1, 0)
                m = comm.mprobe(1, 0)
                self.assertTrue(m)
                rmsg = m.recv()
                self.assertFalse(m)
                #
                comm.send(smsg, 1, 1)
                m = None
                while not m:
                    m = MPI.Message.iprobe(comm, 1, 1)
                rmsg = m.irecv().wait()
            elif rank == 1:
                m = comm.mprobe(0, 0)
                self.assertTrue(m)
                rmsg = m.recv()
                self.assertFalse(m)
                comm.send(rmsg, 0, 0)
                #
                m = None
                while not m:
                    m = MPI.Message.iprobe(comm, 0, 1)
                rmsg = m.irecv().wait()
                comm.send(smsg, 0, 1)
            else:
                rmsg = smsg

            self.assertEqual(smsg, rmsg)


class TestP2PMatchedSelf(BaseTestP2PMatched, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestP2PMatchedWorld(BaseTestP2PMatched, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


@unittest.skipMPI("openmpi(<1.8.5)", MPI.COMM_WORLD.Get_size() > 1)
class TestP2PMatchedSelfDup(TestP2PMatchedSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


@unittest.skipMPI("openmpi(<1.8.5)", MPI.COMM_WORLD.Get_size() > 1)
class TestP2PMatchedWorldDup(TestP2PMatchedWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------

__doc__ = """Message Passing Interface."""

from mpi4py.libmpi cimport *

include "stdlib.pxi"
include "atimport.pxi"

bootstrap()
initialize()

include "allocate.pxi"
include "asstring.pxi"
include "asfspath.pxi"
include "asbuffer.pxi"
include "asarray.pxi"
include "constant.pxi"
include "objmodel.pxi"
include "attrimpl.pxi"
include "errhimpl.pxi"
include "reqimpl.pxi"
include "opimpl.pxi"
include "bufaimpl.pxi"
include "commimpl.pxi"
include "winimpl.pxi"
include "drepimpl.pxi"
include "msgbuffer.pxi"
include "msgpickle.pxi"
include "CAPI.pxi"

# -----------------------------------------------------------------------------

# Assorted constants
# ------------------

UNDEFINED = MPI_UNDEFINED
#: Undefined integer value

ANY_SOURCE = MPI_ANY_SOURCE
#: Wildcard source value for receives

ANY_TAG = MPI_ANY_TAG
#: Wildcard tag value for receives

PROC_NULL = MPI_PROC_NULL
#: Special process rank for send/receive

ROOT = MPI_ROOT
#: Root process for collective inter-communications

BOTTOM = __BOTTOM__
#: Special address for buffers

IN_PLACE = __IN_PLACE__
#: *In-place* option for collective communications


# Predefined Attribute Keyvals
# ----------------------------

KEYVAL_INVALID    = MPI_KEYVAL_INVALID

TAG_UB            = MPI_TAG_UB
IO                = MPI_IO
WTIME_IS_GLOBAL   = MPI_WTIME_IS_GLOBAL

UNIVERSE_SIZE     = MPI_UNIVERSE_SIZE
APPNUM            = MPI_APPNUM

LASTUSEDCODE      = MPI_LASTUSEDCODE

WIN_BASE          = MPI_WIN_BASE
WIN_SIZE          = MPI_WIN_SIZE
WIN_DISP_UNIT     = MPI_WIN_DISP_UNIT
WIN_CREATE_FLAVOR = MPI_WIN_CREATE_FLAVOR
WIN_FLAVOR        = MPI_WIN_CREATE_FLAVOR
WIN_MODEL         = MPI_WIN_MODEL


include "ErrorCode.pyx"
include "Exception.pyx"
include "Datatype.pyx"
include "Status.pyx"
include "Request.pyx"
include "Message.pyx"
include "Op.pyx"
include "Group.pyx"
include "Info.pyx"
include "Errhandler.pyx"
include "Session.pyx"
include "Comm.pyx"
include "Win.pyx"
include "File.pyx"


# Memory Allocation
# -----------------

def Alloc_mem(Aint size: int, Info info: Info = INFO_NULL) -> buffer:
    """
    Allocate memory for message passing and remote memory access.
    """
    cdef void *base = NULL
    CHKERR( MPI_Alloc_mem(size, info.ob_mpi, &base) )
    return mpibuf(base, size)


def Free_mem(mem: buffer) -> None:
    """
    Free memory allocated with `Alloc_mem`.
    """
    cdef void *base = NULL
    cdef buffer buf = asbuffer(mem, &base, NULL, 1)
    CHKERR( MPI_Free_mem(base) )
    buf.release()


# Initialization and Exit
# -----------------------

def Init() -> None:
    """
    Initialize the MPI execution environment.
    """
    CHKERR( MPI_Init(NULL, NULL) )
    initialize()


def Finalize() -> None:
    """
    Terminate the MPI execution environment.
    """
    finalize()
    CHKERR( MPI_Finalize() )


# Levels of MPI threading support
# -------------------------------

THREAD_SINGLE     = MPI_THREAD_SINGLE
#: Only one thread will execute

THREAD_FUNNELED   = MPI_THREAD_FUNNELED
#: MPI calls are *funneled* to the main thread

THREAD_SERIALIZED = MPI_THREAD_SERIALIZED
#: MPI calls are *serialized*

THREAD_MULTIPLE   = MPI_THREAD_MULTIPLE
#: Multiple threads may call MPI


def Init_thread(int required: int = THREAD_MULTIPLE) -> int:
    """
    Initialize the MPI execution environment.
    """
    cdef int provided = MPI_THREAD_SINGLE
    CHKERR( MPI_Init_thread(NULL, NULL, required, &provided) )
    initialize()
    return provided


def Query_thread() -> int:
    """
    Return the level of thread support provided by the MPI library.
    """
    cdef int provided = MPI_THREAD_SINGLE
    CHKERR( MPI_Query_thread(&provided) )
    return provided


def Is_thread_main() -> bool:
    """
    Indicate whether this thread called `Init` or `Init_thread`.
    """
    cdef int flag = 1
    CHKERR( MPI_Is_thread_main(&flag) )
    return <bint>flag


def Is_initialized() -> bool:
    """
    Indicate whether `Init` has been called.
    """
    cdef int flag = 0
    CHKERR( MPI_Initialized(&flag) )
    return <bint>flag


def Is_finalized() -> bool:
    """
    Indicate whether `Finalize` has completed.
    """
    cdef int flag = 0
    CHKERR( MPI_Finalized(&flag) )
    return <bint>flag


# Implementation Information
# --------------------------

# MPI Version Number
# -----------------

VERSION    = MPI_VERSION
SUBVERSION = MPI_SUBVERSION


def Get_version() -> tuple[int, int]:
    """
    Obtain the version number of the MPI standard.
    """
    cdef int version = 1
    cdef int subversion = 0
    CHKERR( MPI_Get_version(&version, &subversion) )
    return (version, subversion)


def Get_library_version() -> str:
    """
    Obtain the version string of the MPI library.
    """
    cdef char name[MPI_MAX_LIBRARY_VERSION_STRING+1]
    cdef int nlen = 0
    CHKERR( MPI_Get_library_version(name, &nlen) )
    return tompistr(name, nlen)


# MPI ABI
# -------

def Get_abi_version() -> tuple[int, int]:
    """
    Obtain the MPI ABI version number.
    """
    cdef int version = 1
    cdef int subversion = 0
    CHKERR( MPI_Abi_get_version(&version, &subversion) )
    return (version, subversion)


def Get_abi_info() -> Info:
    """
    Obtain information about the MPI ABI.
    """
    cdef Info info = <Info>New(Info)
    CHKERR( MPI_Abi_get_info(&info.ob_mpi) )
    return info


def Get_abi_fortran_info() -> Info:
    """
    Obtain information about the Fortran MPI ABI.
    """
    cdef Info info = <Info>New(Info)
    CHKERR( MPI_Abi_get_fortran_info(&info.ob_mpi) )
    return info


# Environmental Inquires
# ----------------------

def Get_processor_name() -> str:
    """
    Obtain the name of the calling processor.
    """
    cdef char name[MPI_MAX_PROCESSOR_NAME+1]
    cdef int nlen = 0
    CHKERR( MPI_Get_processor_name(name, &nlen) )
    return tompistr(name, nlen)


def Get_hw_resource_info() -> Info:
    """
    Obtain information about the hardware platform of the calling processor.
    """
    cdef Info info = <Info>New(Info)
    CHKERR( MPI_Get_hw_resource_info(&info.ob_mpi) )
    return info  # ~> MPI-4.1


# Timers and Synchronization
# --------------------------

def Wtime() -> float:
    """
    Return an elapsed time on the calling processor.
    """
    return MPI_Wtime()


def Wtick() -> float:
    """
    Return the resolution of `Wtime`.
    """
    return MPI_Wtick()


# Control of Profiling
# --------------------

def Pcontrol(int level: int) -> None:
    """
    Control profiling.
    """
    if level < 0 or level > 2: CHKERR( MPI_ERR_ARG )
    CHKERR( MPI_Pcontrol(level) )


# Maximum string sizes
# --------------------

# MPI-1
MAX_PROCESSOR_NAME = MPI_MAX_PROCESSOR_NAME
MAX_ERROR_STRING   = MPI_MAX_ERROR_STRING
# MPI-2
MAX_PORT_NAME      = MPI_MAX_PORT_NAME
MAX_INFO_KEY       = MPI_MAX_INFO_KEY
MAX_INFO_VAL       = MPI_MAX_INFO_VAL
MAX_OBJECT_NAME    = MPI_MAX_OBJECT_NAME
MAX_DATAREP_STRING = MPI_MAX_DATAREP_STRING
# MPI-3
MAX_LIBRARY_VERSION_STRING = MPI_MAX_LIBRARY_VERSION_STRING
# MPI-4
MAX_PSET_NAME_LEN  = MPI_MAX_PSET_NAME_LEN
MAX_STRINGTAG_LEN  = MPI_MAX_STRINGTAG_LEN

# -----------------------------------------------------------------------------

include "typemap.pxi"
include "typestr.pxi"
include "typedec.pxi"

# -----------------------------------------------------------------------------

cdef extern from * nogil:
    int PyMPI_Get_vendor(const char**, int*, int*, int*)


def get_vendor() -> tuple[str, tuple[int, int, int]]:
    """
    Information about the underlying MPI implementation.

    Returns:
      - string with the name of the MPI implementation.
      - integer 3-tuple version number ``(major, minor, micro)``.
    """
    cdef const char *name=NULL
    cdef int major=0, minor=0, micro=0
    CHKERR( PyMPI_Get_vendor(&name, &major, &minor, &micro) )
    return (mpistr(name), (major, minor, micro))


# -----------------------------------------------------------------------------

cdef inline int _mpi_type(object arg, type cls) except -1:
    if isinstance(arg, type):
        if issubclass(arg, cls): return 1
    else:
        if isinstance(arg, cls): return 1
    return 0


def _sizeof(arg: Any) -> int:
    """
    Size in bytes of the underlying MPI handle.
    """
    if _mpi_type(arg, Status):     return sizeof(MPI_Status)
    if _mpi_type(arg, Datatype):   return sizeof(MPI_Datatype)
    if _mpi_type(arg, Request):    return sizeof(MPI_Request)
    if _mpi_type(arg, Message):    return sizeof(MPI_Message)
    if _mpi_type(arg, Op):         return sizeof(MPI_Op)
    if _mpi_type(arg, Group):      return sizeof(MPI_Group)
    if _mpi_type(arg, Info):       return sizeof(MPI_Info)
    if _mpi_type(arg, Errhandler): return sizeof(MPI_Errhandler)
    if _mpi_type(arg, Session):    return sizeof(MPI_Session)
    if _mpi_type(arg, Comm):       return sizeof(MPI_Comm)
    if _mpi_type(arg, Win):        return sizeof(MPI_Win)
    if _mpi_type(arg, File):       return sizeof(MPI_File)
    raise TypeError("expecting an MPI type or instance")


def _addressof(arg: Any) -> int:
    """
    Memory address of the underlying MPI handle.
    """
    cdef void *ptr = NULL
    if isinstance(arg, Status):
        ptr = <void*>&(<Status>arg).ob_mpi
    elif isinstance(arg, Datatype):
        ptr = <void*>&(<Datatype>arg).ob_mpi
    elif isinstance(arg, Request):
        ptr = <void*>&(<Request>arg).ob_mpi
    elif isinstance(arg, Message):
        ptr = <void*>&(<Message>arg).ob_mpi
    elif isinstance(arg, Op):
        ptr = <void*>&(<Op>arg).ob_mpi
    elif isinstance(arg, Group):
        ptr = <void*>&(<Group>arg).ob_mpi
    elif isinstance(arg, Info):
        ptr = <void*>&(<Info>arg).ob_mpi
    elif isinstance(arg, Errhandler):
        ptr = <void*>&(<Errhandler>arg).ob_mpi
    elif isinstance(arg, Session):
        ptr = <void*>&(<Session>arg).ob_mpi
    elif isinstance(arg, Comm):
        ptr = <void*>&(<Comm>arg).ob_mpi
    elif isinstance(arg, Win):
        ptr = <void*>&(<Win>arg).ob_mpi
    elif isinstance(arg, File):
        ptr = <void*>&(<File>arg).ob_mpi
    else:
        raise TypeError("expecting an MPI instance")
    return PyLong_FromVoidPtr(ptr)


def _handleof(arg: Any) -> int:
    """
    Unsigned integer value with the underlying MPI handle.
    """
    if isinstance(arg, Status):
        return <Py_uintptr_t>&((<Status>arg).ob_mpi)
    elif isinstance(arg, Datatype):
        return <Py_uintptr_t>((<Datatype>arg).ob_mpi)
    elif isinstance(arg, Request):
        return <Py_uintptr_t>((<Request>arg).ob_mpi)
    elif isinstance(arg, Message):
        return <Py_uintptr_t>((<Message>arg).ob_mpi)
    elif isinstance(arg, Op):
        return <Py_uintptr_t>((<Op>arg).ob_mpi)
    elif isinstance(arg, Group):
        return <Py_uintptr_t>((<Group>arg).ob_mpi)
    elif isinstance(arg, Info):
        return <Py_uintptr_t>((<Info>arg).ob_mpi)
    elif isinstance(arg, Errhandler):
        return <Py_uintptr_t>((<Errhandler>arg).ob_mpi)
    elif isinstance(arg, Session):
        return <Py_uintptr_t>((<Session>arg).ob_mpi)
    elif isinstance(arg, Comm):
        return <Py_uintptr_t>((<Comm>arg).ob_mpi)
    elif isinstance(arg, Win):
        return <Py_uintptr_t>((<Win>arg).ob_mpi)
    elif isinstance(arg, File):
        return <Py_uintptr_t>((<File>arg).ob_mpi)
    else:
        raise TypeError("expecting an MPI instance")

# -----------------------------------------------------------------------------



#!/usr/bin/env python
"""
Parallel PI computation using Remote Memory Access (RMA)
within Python objects exposing memory buffers (requires NumPy).

usage::

  $ mpiexec -n <nprocs> python cpi-rma.py
"""

from math import pi as PI

from numpy import array

from mpi4py import MPI


def get_n():
    prompt = "Enter the number of intervals: (0 quits) "
    try:
        n = int(input(prompt))
        if n < 0:
            n = 0
    except Exception:
        n = 0
    return n


def comp_pi(n, myrank=0, nprocs=1):
    h = 1.0 / n
    s = 0.0
    for i in range(myrank + 1, n + 1, nprocs):
        x = h * (i - 0.5)
        s += 4.0 / (1.0 + x**2)
    return s * h


def prn_pi(pi, PI):
    message = "pi is approximately %.16f, error is %.16f"
    print(message % (pi, abs(pi - PI)))


nprocs = MPI.COMM_WORLD.Get_size()
myrank = MPI.COMM_WORLD.Get_rank()

n = array(0, dtype=int)
pi = array(0, dtype=float)
mypi = array(0, dtype=float)

if myrank == 0:
    win_n = MPI.Win.Create(n, comm=MPI.COMM_WORLD)
    win_pi = MPI.Win.Create(pi, comm=MPI.COMM_WORLD)
else:
    win_n = MPI.Win.Create(None, comm=MPI.COMM_WORLD)
    win_pi = MPI.Win.Create(None, comm=MPI.COMM_WORLD)

while True:
    if myrank == 0:
        _n = get_n()
        n.fill(_n)
        pi.fill(0.0)
    win_n.Fence()
    if myrank != 0:
        win_n.Get([n, MPI.INT], 0)
    win_n.Fence()
    if n == 0:
        break
    _mypi = comp_pi(n, myrank, nprocs)
    mypi.fill(_mypi)
    win_pi.Fence()
    win_pi.Accumulate([mypi, MPI.DOUBLE], 0, op=MPI.SUM)
    win_pi.Fence()
    if myrank == 0:
        prn_pi(pi, PI)

win_n.Free()
win_pi.Free()



from mpi4py import MPI

try:
    import numpy
except ImportError:
    raise SystemExit from None

# send a upper triangular matrix

N = 10

a = numpy.empty((N, N), dtype=float, order="c")
b = numpy.zeros((N, N), dtype=float, order="c")
a.flat = numpy.arange(a.size, dtype=float)

# compute start and size of each row
i = numpy.arange(N)
blocklen = N - i
disp = N * i + i

# create datatype for upper triangular part
upper = MPI.DOUBLE.Create_indexed(blocklen, disp)
upper.Commit()

# send and recv matrix
myrank = MPI.COMM_WORLD.Get_rank()
MPI.COMM_WORLD.Sendrecv((a, 1, upper), myrank, 0, [b, 1, upper], myrank, 0)

assert numpy.allclose(numpy.triu(b), numpy.triu(a))
assert numpy.allclose(numpy.tril(b, -1), numpy.zeros((N, N)))

upper.Free()



import itertools
import sys
import warnings

from mpi4py import MPI

try:
    import array
except ImportError:
    array = None

try:
    import numpy
except ImportError:
    numpy = None

try:
    import cupy
except ImportError:
    cupy = None
else:
    cupy_version = tuple(map(int, cupy.__version__.split(".", 2)[:2]))

try:
    import numba
    import numba.cuda
except ImportError:
    numba = None
else:
    numba_version = tuple(map(int, numba.__version__.split(".", 2)[:2]))
    if numba_version < (0, 48):
        try:
            warnings.warn(
                "To test Numba GPU arrays, use Numba v0.48.0+.",
                RuntimeWarning,
                stacklevel=1,
            )
        except RuntimeWarning:
            pass
        del numba_version
        numba = None


__all__ = ["loop", "test"]


def make_typemap(entries):
    return {
        typecode: datatype
        for typecode, datatype in entries
        if datatype != MPI.DATATYPE_NULL
    }


TypeMap = make_typemap([
    ("b", MPI.SIGNED_CHAR),
    ("h", MPI.SHORT),
    ("i", MPI.INT),
    ("l", MPI.LONG),
    ("q", MPI.LONG_LONG),
    ("f", MPI.FLOAT),
    ("d", MPI.DOUBLE),
    ("g", MPI.LONG_DOUBLE),
])

TypeMapBool = make_typemap([
    ("?", MPI.C_BOOL),
])

TypeMapInteger = make_typemap([
    ("b", MPI.SIGNED_CHAR),
    ("h", MPI.SHORT),
    ("i", MPI.INT),
    ("l", MPI.LONG),
    ("q", MPI.LONG_LONG),
])

TypeMapUnsigned = make_typemap([
    ("B", MPI.UNSIGNED_CHAR),
    ("H", MPI.UNSIGNED_SHORT),
    ("I", MPI.UNSIGNED_INT),
    ("L", MPI.UNSIGNED_LONG),
    ("Q", MPI.UNSIGNED_LONG_LONG),
])

TypeMapFloat = make_typemap([
    ("f", MPI.FLOAT),
    ("d", MPI.DOUBLE),
    ("g", MPI.LONG_DOUBLE),
])

TypeMapComplex = make_typemap([
    ("F", MPI.C_FLOAT_COMPLEX),
    ("D", MPI.C_DOUBLE_COMPLEX),
    ("G", MPI.C_LONG_DOUBLE_COMPLEX),
])


ArrayBackends = []


def add_backend(cls):
    ArrayBackends.append(cls)
    return cls


class BaseArray:
    #
    backend = None

    TypeMap = TypeMap.copy()
    TypeMap.pop("g", None)

    def __len__(self):
        return len(self.array)

    def __getitem__(self, i):
        return self.array[i]

    def __setitem__(self, i, v):
        self.array[i] = v

    @property
    def mpidtype(self):
        try:
            return self.TypeMap[self.typecode]
        except KeyError:
            return MPI.DATATYPE_NULL

    def as_raw(self):
        return self.array

    def as_mpi(self):
        return (self.as_raw(), self.mpidtype)

    def as_mpi_c(self, count):
        return (self.as_raw(), count, self.mpidtype)

    def as_mpi_v(self, cnt, dsp):
        return (self.as_raw(), (cnt, dsp), self.mpidtype)


if array is not None:

    def product(seq):
        res = 1
        for s in seq:
            res = res * s
        return res

    def mkshape(shape):
        return tuple(int(s) for s in shape)

    @add_backend
    class ArrayArray(BaseArray):
        backend = "array"

        def __init__(self, arg, typecode, shape=None):
            if isinstance(arg, (int, float)):
                if shape is None:
                    shape = ()
                else:
                    try:
                        shape = mkshape(shape)
                    except TypeError:
                        shape = (int(shape),)
                size = product(shape)
                arg = [arg] * size
            else:
                size = len(arg)
                if shape is None:
                    shape = (size,)
                else:
                    shape = mkshape(shape)
                assert size == product(shape)
            self.array = array.array(typecode, arg)

        @property
        def address(self):
            return self.array.buffer_info()[0]

        @property
        def typecode(self):
            return self.array.typecode

        @property
        def itemsize(self):
            return self.array.itemsize

        @property
        def flat(self):
            return self.array

        @property
        def size(self):
            return self.array.buffer_info()[1]


if numpy is not None:

    @add_backend
    class ArrayNumPy(BaseArray):
        backend = "numpy"

        TypeMap = make_typemap([])
        TypeMap.update(TypeMapBool)
        TypeMap.update(TypeMapInteger)
        TypeMap.update(TypeMapUnsigned)
        TypeMap.update(TypeMapFloat)
        TypeMap.update(TypeMapComplex)

        def __init__(self, arg, typecode, shape=None):
            if isinstance(arg, (int, float, complex)):
                if shape is None:
                    shape = ()
            else:
                if shape is None:
                    shape = len(arg)
            self.array = numpy.zeros(shape, typecode)
            if isinstance(arg, (int, float, complex)):
                arg = numpy.asarray(arg).astype(typecode)
                self.array.fill(arg)
            else:
                arg = numpy.asarray(arg).astype(typecode)
                self.array[...] = arg

        @property
        def address(self):
            return self.array.__array_interface__["data"][0]

        @property
        def typecode(self):
            return self.array.dtype.char

        @property
        def itemsize(self):
            return self.array.itemsize

        @property
        def flat(self):
            return self.array.flat

        @property
        def size(self):
            return self.array.size


try:
    import dlpackimpl as dlpack
except ImportError:
    dlpack = None


class BaseDLPackCPU:
    #
    def __dlpack_device__(self):
        return (dlpack.DLDeviceType.kDLCPU, 0)

    def __dlpack__(self, stream=None):
        assert stream is None
        managed = dlpack.make_dl_managed_tensor(self.array)
        capsule = dlpack.make_py_capsule(managed, owned=True)
        return capsule

    def as_raw(self):
        return self


if dlpack is not None and array is not None:

    @add_backend
    class DLPackArray(BaseDLPackCPU, ArrayArray):
        backend = "dlpack-array"

        def __init__(self, arg, typecode, shape=None):
            super().__init__(arg, typecode, shape)


if dlpack is not None and numpy is not None:

    @add_backend
    class DLPackNumPy(BaseDLPackCPU, ArrayNumPy):
        backend = "dlpack-numpy"

        def __init__(self, arg, typecode, shape=None):
            super().__init__(arg, typecode, shape)


def typestr(typecode, itemsize):
    typestr = ""
    if sys.byteorder == "little":
        typestr += "<"
    if sys.byteorder == "big":
        typestr += ">"
    if typecode == "?":
        typestr += "b"
    if typecode in "bhilq":
        typestr += "i"
    if typecode in "BHILQ":
        typestr += "u"
    if typecode in "fdg":
        typestr += "f"
    if typecode in "FDG":
        typestr += "c"
    typestr += str(itemsize)
    return typestr


class BaseFakeGPUArray:
    #
    def set_interface(self, shape, readonly=False):
        self.__cuda_array_interface__ = {
            "version": 0,
            "data": (self.address, readonly),
            "typestr": typestr(self.typecode, self.itemsize),
            "shape": shape,
        }

    def as_raw(self):
        return self


if array is not None:

    @add_backend
    class FakeGPUArrayBasic(BaseFakeGPUArray, ArrayArray):
        def __init__(self, arg, typecode, shape=None, readonly=False):
            super().__init__(arg, typecode, shape)
            self.set_interface((len(self),), readonly)


if numpy is not None:

    @add_backend
    class FakeGPUArrayNumPy(BaseFakeGPUArray, ArrayNumPy):
        def __init__(self, arg, typecode, shape=None, readonly=False):
            super().__init__(arg, typecode, shape)
            self.set_interface(self.array.shape, readonly)


if cupy is not None:

    @add_backend
    class GPUArrayCuPy(BaseArray):
        backend = "cupy"

        TypeMap = make_typemap([])
        if cupy_version >= (11, 6):
            TypeMap.update(TypeMapBool)
        TypeMap.update(TypeMapInteger)
        TypeMap.update(TypeMapUnsigned)
        TypeMap.update(TypeMapFloat)
        TypeMap.update(TypeMapComplex)
        try:
            cupy.array(0, "g")
        except ValueError:
            TypeMap.pop("g", None)
        try:
            cupy.array(0, "G")
        except ValueError:
            TypeMap.pop("G", None)

        def __init__(self, arg, typecode, shape=None, readonly=False):
            if isinstance(arg, (int, float, complex)):
                if shape is None:
                    shape = ()
            else:
                if shape is None:
                    shape = len(arg)
            self.array = cupy.zeros(shape, typecode)
            if isinstance(arg, (int, float, complex)):
                self.array.fill(arg)
            else:
                self.array[:] = cupy.asarray(arg, typecode)
            del readonly  # self.array.flags.readonly = readonly

        @property
        def address(self):
            return self.array.__cuda_array_interface__["data"][0]

        @property
        def typecode(self):
            return self.array.dtype.char

        @property
        def itemsize(self):
            return self.array.itemsize

        @property
        def flat(self):
            return self.array.ravel()

        @property
        def size(self):
            return self.array.size

        def as_raw(self):
            cupy.cuda.get_current_stream().synchronize()
            return self.array


if cupy is not None:
    # Note: we do not create a BaseDLPackGPU class because each GPU library
    # has its own way to get device ID etc, so we have to reimplement the
    # DLPack support anyway

    @add_backend
    class DLPackCuPy(GPUArrayCuPy):
        backend = "dlpack-cupy"
        has_dlpack = None
        dev_type = None

        def __init__(self, arg, typecode, shape=None):
            super().__init__(arg, typecode, shape)
            self.has_dlpack = hasattr(self.array, "__dlpack_device__")
            # TODO(leofang): test CUDA managed memory?
            if cupy.cuda.runtime.is_hip:
                self.dev_type = dlpack.DLDeviceType.kDLROCM
            else:
                self.dev_type = dlpack.DLDeviceType.kDLCUDA

        def __dlpack_device__(self):
            if self.has_dlpack:
                return self.array.__dlpack_device__()
            else:
                return (self.dev_type, self.array.device.id)

        def __dlpack__(self, stream=None):
            cupy.cuda.get_current_stream().synchronize()
            if self.has_dlpack:
                return self.array.__dlpack__(stream=-1)
            else:
                return self.array.toDlpack()

        def as_raw(self):
            return self


if numba is not None:

    @add_backend
    class GPUArrayNumba(BaseArray):
        backend = "numba"

        TypeMap = make_typemap([])
        TypeMap.update(TypeMapBool)
        TypeMap.update(TypeMapInteger)
        TypeMap.update(TypeMapUnsigned)
        TypeMap.update(TypeMapFloat)
        TypeMap.update(TypeMapComplex)

        # one can allocate arrays with those types,
        # but the Numba compiler doesn't support them...
        TypeMap.pop("g", None)
        TypeMap.pop("G", None)

        def __init__(self, arg, typecode, shape=None, readonly=False):
            if isinstance(arg, (int, float, complex)):
                if shape is None:
                    shape = ()
            else:
                if shape is None:
                    shape = len(arg)
            self.array = numba.cuda.device_array(shape, typecode)
            if isinstance(arg, (int, float, complex)):
                if self.array.size > 0:
                    self.array[:] = arg
            elif arg == [] or arg == ():
                self.array = numba.cuda.device_array(0, typecode)
            else:
                if self.array.size > 0:
                    self.array[:] = numba.cuda.to_device(arg)
            del readonly  # self.array.flags.readonly = readonly

        # def __getitem__(self, i):
        #     if isinstance(i, slice):
        #         return self.array[i]
        #     elif i < self.array.size:
        #         return self.array[i]
        #     else:
        #         raise StopIteration

        @property
        def address(self):
            return self.array.__cuda_array_interface__["data"][0]

        @property
        def typecode(self):
            return self.array.dtype.char

        @property
        def itemsize(self):
            return self.array.dtype.itemsize

        @property
        def flat(self):
            if self.array.ndim <= 1:
                return self.array
            else:
                return self.array.ravel()

        @property
        def size(self):
            return self.array.size

        def as_raw(self):
            # numba by default always runs on the legacy default stream
            numba.cuda.default_stream().synchronize()
            return self.array


def loop(*args):
    loop.array = None
    loop.typecode = None
    for array in ArrayBackends:
        loop.array = array
        for typecode in array.TypeMap:
            loop.typecode = typecode
            if not args:
                yield array, typecode
            else:
                for prod in itertools.product(*args):
                    yield (array, typecode, *prod)
    del loop.array
    del loop.typecode


def test(case, **kargs):
    return case.subTest(
        typecode=loop.typecode,
        backend=loop.array.backend,
        **kargs,
    )


def scalar(arg):
    return loop.array(arg, loop.typecode, 1)[0]



import ctypes
import pathlib

from mpi4py import MPI

_libdir = pathlib.Path(__file__).resolve().parent

if MPI._sizeof(MPI.Comm) == ctypes.sizeof(ctypes.c_int):
    MPI_Comm = ctypes.c_int
else:
    MPI_Comm = ctypes.c_void_p

lib = ctypes.CDLL(_libdir / "libhelloworld.so")
lib.sayhello.restype = None
lib.sayhello.argtypes = [MPI_Comm]


def sayhello(comm):
    comm_c = MPI_Comm(comm.handle)
    lib.sayhello(comm_c)



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Management of MPI worker processes."""
# pylint: disable=broad-except
# pylint: disable=too-many-lines
# pylint: disable=protected-access
# pylint: disable=import-outside-toplevel
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import atexit
import collections
import itertools
import os
import sys
import threading
import time
import traceback
import warnings
import weakref

from .. import MPI
from ..util import pkl5
from ._base import BrokenExecutor

# ---

_tls = threading.local()


def serialized(function):
    def wrapper(*args, **kwargs):
        with serialized.lock:
            return function(*args, **kwargs)

    if serialized.lock is None:
        return function
    return wrapper


serialized.lock = None  # type: ignore[attr-defined]


def setup_mpi_threads():
    with setup_mpi_threads.lock:
        thread_level = setup_mpi_threads.thread_level
        if thread_level is None:
            thread_level = MPI.Query_thread()
            setup_mpi_threads.thread_level = thread_level
            if thread_level < MPI.THREAD_MULTIPLE:
                serialized.lock = threading.Lock()
    if thread_level < MPI.THREAD_SERIALIZED:
        warnings.warn(
            "the level of thread support in MPI "
            "should be at least MPI_THREAD_SERIALIZED",
            RuntimeWarning,
            stacklevel=2,
        )


setup_mpi_threads.lock = threading.Lock()  # type: ignore[attr-defined]
setup_mpi_threads.thread_level = None  # type: ignore[attr-defined]


# ---


class RemoteTraceback(Exception):
    pass


def _unwrap_exc(exc, tb):
    exc.__cause__ = RemoteTraceback(tb)
    return exc


class _ExceptionWrapper(Exception):
    def __reduce__(self):
        return _unwrap_exc, self.args


def _wrap_exc(exc, tb):
    exc.__cause__ = None
    exc.__context__ = None
    exc.__traceback__ = None
    return _ExceptionWrapper(exc, tb)


def _format_exc(exc, comm):
    exc_info = (type(exc), exc, exc.__traceback__)
    tb_lines = traceback.format_exception(*exc_info)
    body = "".join(tb_lines)
    host = MPI.Get_processor_name()
    rank = comm.Get_rank()
    size = comm.Get_size()
    info = f"### Worker {rank} of {size} on {host}\n"
    return f'\n{info}"""\n{body}"""'


if sys.version_info >= (3, 11):

    def sys_exception():
        return sys.exception()

else:  # pragma: no cover

    def sys_exception():
        return sys.exc_info()[1]


def os_environ_get(name, default=None):
    varname = f"MPI4PY_FUTURES_{name}"
    if varname not in os.environ:
        oldname = f"MPI4PY_{name}"
        if oldname in os.environ:  # pragma: no cover
            warnings.warn(
                f"environment variable {oldname} is deprecated, use {varname}",
                DeprecationWarning,
                stacklevel=1,
            )
            return os.environ[oldname]
    return os.environ.get(varname, default)


# ---


BACKOFF = 0.001


def _getopt_backoff(options):
    backoff = options.get("backoff")
    if backoff is None:
        backoff = os_environ_get("BACKOFF", BACKOFF)
    return float(backoff)


class Backoff:
    #
    def __init__(self, seconds=BACKOFF):
        self.tval = 0.0
        self.tmax = max(float(seconds), 0.0)
        self.tmin = self.tmax / (1 << 10)

    def reset(self):
        self.tval = 0.0

    def sleep(self):
        time.sleep(self.tval)
        self.tval = min(self.tmax, max(self.tmin, self.tval * 2))


class TaskQueue(collections.deque):
    put = collections.deque.append
    pop = collections.deque.popleft
    add = collections.deque.appendleft


class WorkerSet(collections.deque):
    add = collections.deque.append
    pop = collections.deque.popleft


THREADS_QUEUES = weakref.WeakKeyDictionary()  # type: weakref.WeakKeyDictionary


def join_threads(threads_queues=THREADS_QUEUES):
    items = list(threads_queues.items())
    for _, queue in items:  # pragma: no cover
        queue.put(None)
    for thread, _ in items:  # pragma: no cover
        thread.join()


if hasattr(threading, "_register_atexit"):
    threading._register_atexit(join_threads)
else:  # pragma: no cover
    atexit.register(join_threads)


class Pool:
    #
    def __init__(self, executor, manager, *args):
        self.size = None
        self.event = threading.Event()
        self.queue = queue = TaskQueue()
        self.exref = weakref.ref(executor, lambda _, q=queue: q.put(None))

        args = (self, executor._options, *args)
        thread = threading.Thread(target=manager, args=args)
        self.thread = thread

        setup_mpi_threads()
        thread.daemon = not hasattr(threading, "_register_atexit")
        thread.start()
        THREADS_QUEUES[thread] = queue

    def wait(self):
        self.event.wait()

    def push(self, item):
        self.queue.put(item)

    def done(self):
        self.queue.put(None)

    def join(self):
        self.thread.join()

    def setup(self, size):
        self.size = size
        self.event.set()
        return self.queue

    def cancel(self, handler=None):
        queue = self.queue
        while True:
            try:
                item = queue.pop()
            except LookupError:
                break
            if item is None:
                queue.put(None)
                break
            future, task = item
            if handler:
                handler(future)
            else:
                future.cancel()
                future.set_running_or_notify_cancel()
            del future, task, item

    def broken(self, message):
        lock = None
        executor = self.exref()
        if executor is not None:
            executor._broken = message
            if not executor._shutdown:
                lock = executor._lock

        def handler(future):
            if future.set_running_or_notify_cancel():
                exception = BrokenExecutor(message)
                future.set_exception(exception)

        self.event.set()
        if lock:
            lock.acquire()
        try:
            self.cancel(handler)
        finally:
            if lock:
                lock.release()


def initialize(options):
    initializer = options.pop("initializer", None)
    initargs = options.pop("initargs", ())
    initkwargs = options.pop("initkwargs", {})
    if initializer is not None:
        try:
            initializer(*initargs, **initkwargs)
        except BaseException:
            return False
        else:
            return True
    return True


def _manager_thread(pool, options):
    size = options.pop("max_workers", 1)
    queue = pool.setup(size)
    threads = collections.deque()
    max_threads = size - 1

    def init():
        if not initialize(options):
            pool.broken("initializer failed")
            return False
        return True

    def adjust():
        if len(threads) < max_threads:
            thread = threading.Thread(target=worker)
            thread.start()
            threads.append(thread)

    def execute(future, task):
        func, args, kwargs = task
        result = exception = None
        try:
            result = func(*args, **kwargs)
            future.set_result(result)
        except BaseException:
            exception = sys_exception()
            future.set_exception(exception)
        del result, exception
        del func, args, kwargs
        del future, task

    def finalize():
        for thread in threads:
            thread.join()
        queue.pop()

    def worker():
        backoff = Backoff(_getopt_backoff(options))
        if not init():
            queue.put(None)
            return
        while True:
            try:
                item = queue.pop()
                backoff.reset()
            except LookupError:
                backoff.sleep()
                continue
            if item is None:
                queue.put(None)
                break
            future, task = item
            if future.set_running_or_notify_cancel():
                if queue:
                    adjust()
                execute(future, task)
            del future, task, item

    worker()
    finalize()


def _manager_comm(pool, options, comm, sync=True):
    assert comm != MPI.COMM_NULL  # noqa: S101
    assert comm.Is_inter()  # noqa: S101
    assert comm.Get_size() == 1  # noqa: S101
    comm = client_sync(comm, options, sync)
    if not client_init(comm, options):
        pool.broken("initializer failed")
        client_stop(comm)
        return
    size = comm.Get_remote_size()
    queue = pool.setup(size)
    workers = WorkerSet(range(size))
    client_exec(comm, options, 0, workers, queue)
    client_stop(comm)


def _manager_spawn(pool, options):
    pyexe = options.pop("python_exe", None)
    pyargs = options.pop("python_args", None)
    nprocs = options.pop("max_workers", None)
    info = options.pop("mpi_info", None)
    comm = serialized(client_spawn)(pyexe, pyargs, nprocs, info)
    _manager_comm(pool, options, comm)


def _manager_service(pool, options):
    service = options.pop("service", None)
    info = options.pop("mpi_info", None)
    comm = serialized(client_connect)(service, info)
    _manager_comm(pool, options, comm)


def ThreadPool(executor):
    # pylint: disable=invalid-name
    return Pool(executor, _manager_thread)


def SpawnPool(executor):
    # pylint: disable=invalid-name
    return Pool(executor, _manager_spawn)


def ServicePool(executor):
    # pylint: disable=invalid-name
    return Pool(executor, _manager_service)


def WorkerPool(executor):
    # pylint: disable=invalid-name
    if SharedPool is not None:
        return SharedPool(executor)
    if "service" in executor._options:
        return ServicePool(executor)
    return SpawnPool(executor)


# ---

SharedPool = None  # pylint: disable=invalid-name


def _set_shared_pool(obj):
    # pylint: disable=global-statement
    global SharedPool
    SharedPool = obj


class SharedPoolCtx:
    #
    def __init__(self):
        self.lock = threading.Lock()
        self.comm = MPI.COMM_NULL
        self.intracomm = MPI.COMM_NULL
        self.on_root = None
        self.counter = None
        self.workers = None
        self.threads = weakref.WeakKeyDictionary()

    def __reduce__(self):
        return "SharedPool"

    def _initialize_remote(self):
        barrier(self.intracomm)
        server_init(self.comm)

    def _initialize(self, options, tag):
        if tag == 0:
            self.comm = client_sync(self.comm, options)
            return client_init(self.comm, options)
        if options.get("initializer") is None:
            return True
        task = (self._initialize_remote, (), {})
        reqs = isendtoall(self.comm, task, tag)
        waitall(self.comm, reqs, poll=True)
        success = client_init(self.comm, options)
        recvfromall(self.comm, tag)
        return success

    def _manager(self, pool, options):
        if self.counter is None:
            options["max_workers"] = 1
            set_comm_server(MPI.COMM_SELF)
            _manager_thread(pool, options)
            return
        with self.lock:
            tag = next(self.counter)
            if not self._initialize(options, tag):
                pool.broken("initializer failed")
                return
        comm = self.comm
        size = comm.Get_remote_size()
        queue = pool.setup(size)
        client_exec(comm, options, tag, self.workers, queue)

    def __call__(self, executor):
        assert SharedPool is self  # noqa: S101
        with self.lock:
            pool = Pool(executor, self._manager)
            del THREADS_QUEUES[pool.thread]
            self.threads[pool.thread] = pool.queue
        return pool

    def __enter__(self):
        assert SharedPool is None  # noqa: S101
        comm, root = MPI.COMM_WORLD, 0
        self.on_root = comm.Get_rank() == root
        self.comm, self.intracomm = comm_split(comm, root)
        if self.comm != MPI.COMM_NULL and self.on_root:
            size = self.comm.Get_remote_size()
            self.counter = itertools.count(0)
            self.workers = WorkerSet(range(size))
        _set_shared_pool(self)
        return self if self.on_root else None

    def __exit__(self, *args):
        assert SharedPool is self  # noqa: S101
        if self.on_root:
            join_threads(self.threads)
        if self.comm != MPI.COMM_NULL:
            comm = self.comm
            if self.on_root:
                if next(self.counter) == 0:
                    options = {"main": False}
                    self._initialize(options, 0)
                client_stop(comm)
            else:
                intracomm = self.intracomm
                set_comm_server(intracomm)
                server_main_comm(comm)
                del_comm_server()
                intracomm.Free()
        if not self.on_root:
            join_threads(self.threads)
        _set_shared_pool(None)
        self.comm = MPI.COMM_NULL
        self.intracomm = MPI.COMM_NULL
        self.on_root = None
        self.counter = None
        self.workers = None
        self.threads.clear()
        return False


# ---


def comm_split(comm, root):
    if comm.Get_size() == 1:
        comm = MPI.Intercomm(MPI.COMM_NULL)
        intracomm = MPI.Intracomm(MPI.COMM_NULL)
        return comm, intracomm
    rank = comm.Get_rank()
    if MPI.Get_version() >= (2, 2):
        allgroup = comm.Get_group()
        if rank == root:
            group = allgroup.Incl([root])
        else:
            group = allgroup.Excl([root])
        allgroup.Free()
        intracomm = comm.Create(group)
        group.Free()
    else:  # pragma: no cover
        color = 0 if rank == root else 1
        intracomm = comm.Split(color, key=0)

    if rank == root:
        local_leader = 0
        remote_leader = 0 if root else 1
    else:
        local_leader = 0
        remote_leader = root
    intercomm = intracomm.Create_intercomm(
        local_leader, comm, remote_leader, tag=0
    )
    if rank == root:
        intracomm.Free()
    return intercomm, intracomm


# ---


def _comm_executor_helper(executor, comm, root):
    def _manager(pool, options, comm, root):
        if comm.Get_size() == 1:
            options["max_workers"] = 1
            set_comm_server(MPI.COMM_SELF)
            _manager_thread(pool, options)
            return
        comm, _ = serialized(comm_split)(comm, root)
        _manager_comm(pool, options, comm, sync=False)

    if comm.Get_rank() == root:
        if SharedPool is not None:
            pool = SharedPool(executor)
        else:
            pool = Pool(executor, _manager, comm, root)
        executor._pool = pool
    else:
        comm, intracomm = comm_split(comm, root)
        set_comm_server(intracomm)
        server_main_comm(comm, sync=False)
        del_comm_server()
        intracomm.Free()


# ---


def _getenv_use_pkl5():
    value = os_environ_get("USE_PKL5")
    if value is None:
        return None
    if value.lower() in ("false", "no", "off", "n", "0"):
        return False
    if value.lower() in ("true", "yes", "on", "y", "1"):
        return True
    warnings.warn(
        f"environment variable MPI4PY_FUTURES_USE_PKL5: "
        f"unexpected value {value!r}",
        RuntimeWarning,
        stacklevel=1,
    )
    return False


def _setopt_use_pkl5(options):
    use_pkl5 = options.get("use_pkl5")
    if use_pkl5 is None:
        use_pkl5 = _getenv_use_pkl5()
    if use_pkl5 is not None:
        options["use_pkl5"] = use_pkl5


def _get_comm(comm, options):
    use_pkl5 = options.pop("use_pkl5", None)
    if use_pkl5:
        return pkl5.Intercomm(comm)
    return comm


def _get_request(comm):
    use_pkl5 = isinstance(comm, pkl5.Comm)
    if use_pkl5:
        return pkl5.Request
    return MPI.Request


# ---


def barrier(comm):
    try:
        request = comm.Ibarrier()
        backoff = Backoff()
        while not request.Test():
            backoff.sleep()
    except (NotImplementedError, MPI.Exception):  # pragma: no cover
        buf = [None, 0, MPI.BYTE]
        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB)
        sendreqs, recvreqs = [], []
        if comm.Is_inter():
            size = comm.Get_remote_size()
        else:
            size = comm.Get_size()
        for pid in range(size):
            recvreqs.append(comm.Irecv(buf, pid, tag))
            sendreqs.append(comm.Issend(buf, pid, tag))
        backoff = Backoff()
        while not MPI.Request.Testall(recvreqs):
            backoff.sleep()
        MPI.Request.Waitall(sendreqs)


def bcast_send(comm, data):
    if MPI.VERSION >= 2:
        comm.bcast(data, MPI.ROOT)
    else:  # pragma: no cover
        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB)
        sendtoall(comm, data, tag)


def bcast_recv(comm):
    if MPI.VERSION >= 2:
        data = comm.bcast(None, 0)
    else:  # pragma: no cover
        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB)
        data = comm.recv(None, 0, tag)
    return data


def isendtoall(comm, data, tag=0):
    size = comm.Get_remote_size()
    return [comm.issend(data, pid, tag) for pid in range(size)]


def waitall(comm, requests, poll=False):
    if poll:
        request_testall = _get_request(comm).testall
        backoff = Backoff()
        while True:
            done, objs = request_testall(requests)
            if done:
                return objs
            backoff.sleep()
    else:
        request_waitall = _get_request(comm).waitall
        return request_waitall(requests)


def sendtoall(comm, data, tag=0):
    requests = isendtoall(comm, data, tag)
    waitall(comm, requests)


def recvfromall(comm, tag=0):
    size = comm.Get_remote_size()
    return [comm.recv(None, pid, tag) for pid in range(size)]


def disconnect(comm):
    try:
        comm.Disconnect()
    except NotImplementedError:  # pragma: no cover
        comm.Free()


# ---


def client_sync(comm, options, sync=True):
    serialized(barrier)(comm)
    _setopt_use_pkl5(options)
    if sync:
        options = _sync_get_data(options)
    serialized(bcast_send)(comm, options)
    comm = _get_comm(comm, options)
    return comm


def client_init(comm, options):
    serialized(bcast_send)(comm, _init_get_data(options))
    sbuf = bytearray([False])
    rbuf = bytearray([False])
    serialized(comm.Allreduce)(sbuf, rbuf, op=MPI.LAND)
    success = bool(rbuf[0])
    return success


def client_exec(comm, options, tag, worker_set, task_queue):
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    backoff = Backoff(_getopt_backoff(options))

    status = MPI.Status()
    comm_recv = serialized(comm.recv)
    comm_isend = serialized(comm.issend)
    comm_iprobe = serialized(comm.iprobe)
    request_free = serialized(_get_request(comm).Free)

    pending = {}

    def iprobe():
        pid = MPI.ANY_SOURCE
        return comm_iprobe(pid, tag, status)

    def probe():
        pid = MPI.ANY_SOURCE
        backoff.reset()
        while not comm_iprobe(pid, tag, status):
            backoff.sleep()

    def recv():
        pid = MPI.ANY_SOURCE
        try:
            task = comm_recv(None, pid, tag, status)
        except BaseException:
            task = (None, sys_exception())
        pid = status.source
        worker_set.add(pid)

        future, request = pending.pop(pid)
        request_free(request)
        result, exception = task
        if exception is None:
            future.set_result(result)
        else:
            future.set_exception(exception)

        del result, exception
        del future, task

    def send():
        try:
            pid = worker_set.pop()
        except LookupError:  # pragma: no cover
            return False

        try:
            item = task_queue.pop()
        except LookupError:  # pragma: no cover
            worker_set.add(pid)
            return False

        if item is None:
            worker_set.add(pid)
            return True

        future, task = item
        if not future.set_running_or_notify_cancel():
            worker_set.add(pid)
            return False

        try:
            request = comm_isend(task, pid, tag)
            pending[pid] = (future, request)
        except BaseException:
            worker_set.add(pid)
            future.set_exception(sys_exception())

        del future, task, item
        return None

    while True:
        if task_queue and worker_set:
            backoff.reset()
            stop = send()
            if stop:
                break
        if pending and iprobe():
            backoff.reset()
            recv()
        backoff.sleep()
    while pending:
        probe()
        recv()


def client_stop(comm):
    serialized(sendtoall)(comm, None)
    serialized(disconnect)(comm)


def server_sync(comm, sync=True):
    barrier(comm)
    options = bcast_recv(comm)
    if sync:
        options = _sync_set_data(options)
    comm = _get_comm(comm, options)
    return comm, options


def server_init(comm):
    options = bcast_recv(comm)
    success = initialize(options)
    sbuf = bytearray([success])
    rbuf = bytearray([True])
    comm.Allreduce(sbuf, rbuf, op=MPI.LAND)
    return success


def server_exec(comm, options):
    backoff = Backoff(_getopt_backoff(options))

    status = MPI.Status()
    comm_recv = comm.recv
    comm_isend = comm.issend
    comm_iprobe = comm.iprobe
    request_test = _get_request(comm).test

    def exception():
        exc = sys_exception()
        tb = _format_exc(exc, comm)
        return _wrap_exc(exc, tb)

    def recv():
        pid, tag = MPI.ANY_SOURCE, MPI.ANY_TAG
        backoff.reset()
        while not comm_iprobe(pid, tag, status):
            backoff.sleep()
        pid, tag = status.source, status.tag
        try:
            task = comm_recv(None, pid, tag, status)
        except BaseException:
            task = exception()
        return task

    def call(task):
        if isinstance(task, BaseException):
            return (None, task)
        func, args, kwargs = task
        try:
            result = func(*args, **kwargs)
        except BaseException:
            return (None, exception())
        else:
            return (result, None)

    def send(task):
        pid, tag = status.source, status.tag
        try:
            request = comm_isend(task, pid, tag)
        except BaseException:
            task = (None, exception())
            request = comm_isend(task, pid, tag)
        backoff.reset()
        while not request_test(request)[0]:
            backoff.sleep()

    while True:
        task = recv()
        if task is None:
            break
        task = call(task)
        send(task)


def server_stop(comm):
    disconnect(comm)


# ---

MAIN_RUN_NAME = "__worker__"


def import_main(mod_name, mod_path, init_globals, run_name):
    import runpy
    import types

    module = types.ModuleType(run_name)
    if init_globals is not None:
        module.__dict__.update(init_globals)
        module.__name__ = run_name

    class TempModulePatch(runpy._TempModule):
        # pylint: disable=too-few-public-methods
        def __init__(self, mod_name):
            super().__init__(mod_name)
            self.module = module

    TempModule = runpy._TempModule  # pylint: disable=invalid-name
    runpy._TempModule = TempModulePatch
    import_main.sentinel = (mod_name, mod_path)
    main_module = sys.modules["__main__"]
    try:
        sys.modules["__main__"] = sys.modules[run_name] = module
        if mod_name:  # pragma: no cover
            runpy.run_module(mod_name, run_name=run_name, alter_sys=True)
        elif mod_path:  # pragma: no branch
            safe_path = getattr(sys.flags, "safe_path", sys.flags.isolated)
            if not safe_path:  # pragma: no branch
                dir_path = os.path.dirname(mod_path)  # noqa: PTH120
                sys.path[0] = os.path.realpath(dir_path)
            runpy.run_path(mod_path, run_name=run_name)
        sys.modules["__main__"] = sys.modules[run_name] = module
    except BaseException:  # pragma: no cover
        sys.modules["__main__"] = main_module
        raise
    finally:
        del import_main.sentinel
        runpy._TempModule = TempModule


def _sync_get_data(options):
    main = sys.modules["__main__"]
    sys.modules.setdefault(MAIN_RUN_NAME, main)
    import_main_module = options.pop("main", True)

    data = options.copy()
    data.pop("initializer", None)
    data.pop("initargs", None)
    data.pop("initkwargs", None)

    if import_main_module:
        spec = getattr(main, "__spec__", None)
        name = getattr(spec, "name", None)
        path = getattr(main, "__file__", None)
        if name is not None:  # pragma: no cover
            data["@main:mod_name"] = name
        if path is not None:  # pragma: no branch
            data["@main:mod_path"] = path

    return data


def _sync_set_data(data):
    if "path" in data:
        sys.path.extend(data.pop("path"))
    if "wdir" in data:
        os.chdir(data.pop("wdir"))
    if "env" in data:
        os.environ.update(data.pop("env"))

    mod_name = data.pop("@main:mod_name", None)
    mod_path = data.pop("@main:mod_path", None)
    mod_glbs = data.pop("globals", None)
    import_main(mod_name, mod_path, mod_glbs, MAIN_RUN_NAME)

    return data


def _init_get_data(options):
    keys = ("initializer", "initargs", "initkwargs")
    vals = (None, (), {})
    data = {k: options.pop(k, v) for k, v in zip(keys, vals)}
    return data


# ---


def _check_recursive_spawn():  # pragma: no cover
    if not hasattr(import_main, "sentinel"):
        return
    main_name, main_path = import_main.sentinel
    main_info = "\n"
    if main_name is not None:
        main_info += f"    main name: {main_name!r}\n"
    if main_path is not None:
        main_info += f"    main path: {main_path!r}\n"
    main_info += "\n"
    sys.stderr.write(
        """
    The main script or module attempted to spawn new MPI worker processes.
    This probably means that you have forgotten to use the proper idiom in
    your main script or module:

        if __name__ == '__main__':
            ...

    This error is unrecoverable. The MPI execution environment had to be
    aborted. The name/path of the offending main script/module follows:
    """
        + main_info
    )
    sys.stderr.flush()
    time.sleep(1)
    MPI.COMM_WORLD.Abort(1)


FLAG_OPT_MAP = {
    "debug": "d",
    "inspect": "i",
    "interactive": "i",
    "optimize": "O",
    "dont_write_bytecode": "B",
    "no_user_site": "s",
    "no_site": "S",
    "ignore_environment": "E",
    "verbose": "v",
    "bytes_warning": "b",
    "quiet": "q",
    "hash_randomization": "R",
    "isolated": "I",
    # 'dev_mode': 'Xdev',
    # 'utf8_mode': 'Xutf8',
    # 'warn_default_encoding': 'Xwarn_default_encoding',
    "safe_path": "P",
    # 'int_max_str_digits': 'Xint_max_str_digits=0'
}


def get_python_flags():
    args = []
    for flag, opt in FLAG_OPT_MAP.items():
        val = getattr(sys.flags, flag, 0)
        val = val if opt[0] != "i" else 0
        val = val if opt[0] != "Q" else min(val, 1)
        if val > 0:
            args.append("-" + opt * val)
    for opt in sys.warnoptions:  # pragma: no cover
        args.append("-W" + opt)
    sys_xoptions = getattr(sys, "_xoptions", {})
    for opt, val in sys_xoptions.items():  # pragma: no cover
        args.append("-X" + opt if val is True else "-X" + opt + "=" + val)
    return args


def get_max_workers():
    max_workers = os_environ_get("MAX_WORKERS")
    if max_workers is not None:
        return int(max_workers)
    if MPI.UNIVERSE_SIZE != MPI.KEYVAL_INVALID:  # pragma: no branch
        universe_size = MPI.COMM_WORLD.Get_attr(MPI.UNIVERSE_SIZE)
        if universe_size is not None:  # pragma: no cover
            world_size = MPI.COMM_WORLD.Get_size()
            return max(universe_size - world_size, 1)
    return 1


def get_spawn_module():
    return __spec__.parent + ".server"


def client_spawn(
    python_exe=None,
    python_args=None,
    max_workers=None,
    mpi_info=None,
):
    _check_recursive_spawn()
    if python_exe is None:
        python_exe = sys.executable
    if python_args is None:
        python_args = []
    if max_workers is None:
        max_workers = get_max_workers()
    if mpi_info is None:
        mpi_info = {"soft": f"1:{max_workers}"}

    args = get_python_flags() + list(python_args)
    args.extend(["-m", get_spawn_module()])
    info = MPI.Info.Create()
    info.update(mpi_info)
    comm = MPI.COMM_SELF.Spawn(python_exe, args, max_workers, info)
    info.Free()
    return comm


# ---


SERVICE = __spec__.parent
SERVER_HOST = "localhost"
SERVER_BIND = ""
SERVER_PORT = 31415


def get_service():
    return os_environ_get("SERVICE", SERVICE)


def get_server_host():
    return os_environ_get("SERVER_HOST", SERVER_HOST)


def get_server_bind():
    return os_environ_get("SERVER_BIND", SERVER_BIND)


def get_server_port():
    return int(os_environ_get("SERVER_PORT", SERVER_PORT))


def client_lookup(address):
    from socket import socket

    host, port = address
    host = host or get_server_host()
    port = port or get_server_port()
    address = (host, int(port))
    sock = socket()
    sock.connect(address)
    try:
        fdes = sock.fileno()
        peer = MPI.Comm.Join(fdes)
    finally:
        sock.close()
    mpi_port = peer.recv(None, 0)
    peer.Disconnect()
    return mpi_port


def server_publish(address, mpi_port):
    from socket import SO_REUSEADDR, SOL_SOCKET, socket

    host, port = address
    host = host or get_server_bind()
    port = port or get_server_port()
    address = (host, int(port))
    serversock = socket()
    serversock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    serversock.bind(address)
    serversock.listen(0)
    try:
        sock = serversock.accept()[0]
    finally:
        serversock.close()
    try:
        fdes = sock.fileno()
        peer = MPI.Comm.Join(fdes)
    finally:
        sock.close()
    peer.send(mpi_port, 0)
    peer.Disconnect()


def client_connect(service, mpi_info=None):
    info = MPI.INFO_NULL
    if mpi_info:
        info = MPI.Info.Create()
        info.update(mpi_info)
    if not isinstance(service, (list, tuple)):
        service = service or get_service()
        port = MPI.Lookup_name(service, info)
    else:
        port = client_lookup(service)

    comm = MPI.COMM_SELF.Connect(port, info, root=0)
    if info != MPI.INFO_NULL:
        info.Free()
    return comm


def server_accept(
    service,
    mpi_info=None,
    comm=MPI.COMM_WORLD,
    root=0,
):
    info = MPI.INFO_NULL
    if comm.Get_rank() == root:
        if mpi_info:
            info = MPI.Info.Create()
            info.update(mpi_info)
    port = None
    if comm.Get_rank() == root:
        port = MPI.Open_port(info)
    if comm.Get_rank() == root:
        if not isinstance(service, (list, tuple)):
            service = service or get_service()
            MPI.Publish_name(service, port, info)
        else:
            server_publish(service, port)
            service = None

    comm = comm.Accept(port, info, root)
    if port is not None:
        if service is not None:
            MPI.Unpublish_name(service, port, info)
        MPI.Close_port(port)
    if info != MPI.INFO_NULL:
        info.Free()
    return comm


# ---


def get_comm_server():
    try:
        return _tls.comm_server
    except AttributeError:
        raise RuntimeError(
            "communicator is not accessible",
        ) from None


def set_comm_server(intracomm):
    _tls.comm_server = intracomm


def del_comm_server():
    del _tls.comm_server


def server_main_comm(comm, sync=True):
    assert comm != MPI.COMM_NULL  # noqa: S101
    assert comm.Is_inter()  # noqa: S101
    assert comm.Get_remote_size() == 1  # noqa: S101
    comm, options = server_sync(comm, sync)
    server_init(comm)
    server_exec(comm, options)
    server_stop(comm)


def server_main_spawn():
    comm = MPI.Comm.Get_parent()
    set_comm_server(MPI.COMM_WORLD)
    server_main_comm(comm)
    del_comm_server()


def server_main_service():
    from getopt import getopt

    longopts = ["bind=", "port=", "service=", "info="]
    optlist, _ = getopt(sys.argv[1:], "", longopts)
    optdict = {opt[2:]: val for opt, val in optlist}

    if "bind" in optdict or "port" in optdict:
        bind = optdict.get("bind") or get_server_bind()
        port = optdict.get("port") or get_server_port()
        service = (bind, int(port))
    else:
        service = optdict.get("service") or get_service()
    info = optdict.get("info", "").split(",")
    info = dict(k_v.split("=", 1) for k_v in info if k_v)

    comm = server_accept(service, info)
    set_comm_server(MPI.COMM_WORLD)
    server_main_comm(comm)
    del_comm_server()


def server_main():
    from ..run import set_abort_status

    try:
        comm = MPI.Comm.Get_parent()
        if comm != MPI.COMM_NULL:
            server_main_spawn()
        else:
            server_main_service()
    except BaseException:
        set_abort_status(1)
        raise


# ---



import sys
import textwrap

import mpiunittest as unittest

from mpi4py import MPI

ModuleType = type(MPI)
ClassType = type(MPI.Comm)
FunctionType = type(MPI.Init)
StaticMethodType = type(MPI.buffer.allocate)
ClassMethodType = type(MPI.Comm.Get_parent)
MethodDescrType = type(MPI.Comm.Get_rank)
GetSetDescrType = type(MPI.Comm.rank)


def getdocstr(mc, docstrings, namespace=None):
    name = getattr(mc, "__name__", None)
    if name is None:
        return
    if name in ("__builtin__", "builtins"):
        return
    if namespace:
        name = f"{namespace}.{name}"
    if type(mc) in (
        ModuleType,
        ClassType,
    ):
        doc = getattr(mc, "__doc__", None)
        if doc == "<undocumented>":
            return
        docstrings[name] = doc
        for _k, v in vars(mc).items():
            if isinstance(v, (classmethod, staticmethod)):
                v = v.__get__(mc)
            getdocstr(v, docstrings, name)
    elif type(mc) in (
        FunctionType,
        StaticMethodType,
        ClassMethodType,
        MethodDescrType,
        GetSetDescrType,
    ):
        doc = getattr(mc, "__doc__", None)
        if doc == "<undocumented>":
            return
        if doc is not None:
            _, nl, doc = doc.partition("\n")
            if nl and doc.startswith("\n"):
                doc = textwrap.dedent(doc[1:])
        docstrings[name] = doc


@unittest.skipIf(hasattr(sys, "pypy_version_info"), "pypy")
class TestDoc(unittest.TestCase):
    #
    def testDoc(self):
        ignore = {"py2f", "f2py"}
        invalid = False
        missing = False
        docs = {}
        getdocstr(MPI, docs)
        for k in docs:
            doc = docs[k]
            name = k.split(".")[-1]
            if name in ignore:
                continue
            if not doc and name.startswith("_"):
                continue
            if doc is None:
                print(f"'{k}': missing docstring")
                missing = True
                continue
            if not doc.strip():
                print(f"'{k}': empty docstring")
                missing = True
                continue
            if doc.startswith("\n") and not doc.endswith(" "):
                print(f"'{k}': mismatch start and end whitespace")
                invalid = True
            if not doc.startswith("\n") and doc.endswith(" "):
                print(f"'{k}': mismatch start and end whitespace")
                invalid = True
            if doc.replace(" ", "").endswith("\n\n"):
                print(f"'{k}': docstring ends with too many newlines")
                invalid = True
            doc = doc.strip()
            if doc[0] == doc[0].lower():
                print(f"'{k}': docstring starts with lowercase")
                invalid = True
            if not doc.endswith("."):
                print(f"'{k}': docstring does not end with '.'")
                invalid = True
            summary, _, _ = doc.partition("\n")
            if not summary.endswith("."):
                print(f"'{k}': summary line does not end with '.'")
                invalid = True
        self.assertFalse(missing)
        self.assertFalse(invalid)


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    int PyIndex_Check(object)
    int PySequence_Check(object)
    object PyNumber_Index(object)
    Py_ssize_t PySequence_Size(object) except -1

cdef inline int is_integral(object ob) noexcept:
    if not PyIndex_Check(ob):    return 0
    if not PySequence_Check(ob): return 1
    try: PySequence_Size(ob)
    except: pass  # noqa
    else: return 0
    try: PyNumber_Index(ob)
    except: return 0  # noqa
    else: return 1

# -----------------------------------------------------------------------------

cdef extern from * nogil:
    const int INT_MAX

ctypedef fused count_t:
    int
    MPI_Count

ctypedef fused integral_t:
    int
    MPI_Aint
    MPI_Count

cdef inline int chklength(Py_ssize_t size) except -1:
    cdef int overflow = (size > (<Py_ssize_t>INT_MAX))
    if overflow: raise OverflowError("length {size} larger than {INT_MAX}")
    return 0

cdef inline object newarray(Py_ssize_t n, integral_t **p):
    return allocate(n, sizeof(integral_t), p)

cdef inline object getarray(object ob, count_t *n, integral_t **p):
    cdef Py_ssize_t size = len(ob)
    if count_t is int: chklength(size)
    cdef integral_t *base = NULL
    cdef object mem = newarray(size, &base)
    for i in range(size):
        base[i] = PyNumber_Index(ob[i])
    n[0] = <count_t> size
    p[0] = base
    return mem

cdef inline object chkarray(object ob, count_t n, integral_t **p):
    cdef count_t size = 0
    cdef object mem = getarray(ob, &size, p)
    if n != size:
        raise ValueError(f"expecting {n} items, got {size}")
    return mem

# -----------------------------------------------------------------------------

cdef inline object asarray_Datatype(
    object sequence, MPI_Count size, MPI_Datatype **p,
):
    cdef MPI_Datatype *array = NULL
    if size != len(sequence):
        raise ValueError(f"expecting {size} items, got {len(sequence)}")
    cdef object ob = allocate(size, sizeof(MPI_Datatype), &array)
    for i in range(size):
        array[i] = (<Datatype?>sequence[i]).ob_mpi
    p[0] = array
    return ob

cdef inline object asarray_Info(
    object sequence, MPI_Count size, MPI_Info **p,
):
    cdef MPI_Info *array = NULL
    cdef MPI_Info info = MPI_INFO_NULL
    cdef object ob
    if sequence is None or isinstance(sequence, Info):
        if sequence is not None:
            info = (<Info?>sequence).ob_mpi
        ob = allocate(size, sizeof(MPI_Info), &array)
        for i in range(size):
            array[i] = info
    else:
        if size != len(sequence):
            raise ValueError(f"expecting {size} items, got {len(sequence)}")
        ob = allocate(size, sizeof(MPI_Datatype), &array)
        for i in range(size):
            array[i] = (<Info?>sequence[i]).ob_mpi
    p[0] = array
    return ob

# -----------------------------------------------------------------------------

cdef inline int is_string(object obj):
    return isinstance(obj, str) or isinstance(obj, bytes)

cdef inline int is_fspath(object obj):
    return hasattr(obj, "__fspath__")

cdef inline object asstring(object ob, char *s[]):
    cdef Py_ssize_t n = 0
    cdef char *p = NULL, *q = NULL
    ob = asmpistr(ob, &p)
    PyBytes_AsStringAndSize(ob, &p, &n)
    cdef object mem = allocate(n+1, sizeof(char), &q)
    <void>memcpy(q, p, <size_t>n)
    q[n] = 0
    s[0] = q
    return mem

cdef inline object asfspath(object ob, char *s[]):
    cdef Py_ssize_t n = 0
    cdef char *p = NULL, *q = NULL
    ob = asmpifspath(ob, &p)
    PyBytes_AsStringAndSize(ob, &p, &n)
    cdef object mem = allocate(n+1, sizeof(char), &q)
    <void>memcpy(q, p, <size_t>n)
    q[n] = 0
    s[0] = q
    return mem

cdef inline object asarray_string(object sequence, char ***p):
    cdef char** array = NULL
    cdef Py_ssize_t size = len(sequence)
    cdef object ob = allocate(size+1, sizeof(char*), &array)
    for i in range(size):
        sequence[i] = asstring(sequence[i], &array[i])
    array[size] = NULL
    p[0] = array
    return (sequence, ob)

cdef inline object asarray_fspath(object sequence, char ***p):
    cdef char** array = NULL
    cdef Py_ssize_t size = len(sequence)
    cdef object ob = allocate(size+1, sizeof(char*), &array)
    for i in range(size):
        sequence[i] = asfspath(sequence[i], &array[i])
    array[size] = NULL
    p[0] = array
    return (sequence, ob)

cdef inline object asarray_argv(object sequence, char ***p):
    if sequence is None:
        p[0] = MPI_ARGV_NULL
        return None
    if is_string(sequence):
        sequence = [sequence]
    else:
        sequence = list(sequence)
    return asarray_string(sequence, p)

cdef inline object asarray_cmds(object sequence, int *count, char ***p):
    if is_string(sequence) or is_fspath(sequence):
        raise ValueError("expecting a sequence of path-like")
    sequence = list(sequence)
    count[0] = <int>len(sequence)
    return asarray_fspath(sequence, p)

cdef inline object asarray_argvs(object sequence, int size, char ****p):
    if sequence is None:
        p[0] = MPI_ARGVS_NULL
        return None
    if is_string(sequence):
        sequence = [sequence] * size
    else:
        sequence = list(sequence)
        if size != len(sequence):
            raise ValueError(f"expecting {size} items, got {len(sequence)}")
    cdef char*** array = NULL
    cdef object ob = allocate(size+1, sizeof(char**), &array)
    cdef object argv
    for i in range(size):
        argv = sequence[i]
        if argv is None: argv = []
        sequence[i] = asarray_argv(argv, &array[i])
    array[size] = NULL
    p[0] = array
    return (sequence, ob)

cdef inline object asarray_nprocs(object sequence, int size, int **p):
    cdef object ob
    cdef int *array = NULL
    cdef int value = 1
    if sequence is None or is_integral(sequence):
        if sequence is not None:
            value = sequence
        ob = newarray(size, &array)
        for i in range(size):
            array[i] = value
    else:
        ob = chkarray(sequence, size, &array)
    p[0] = array
    return ob

# -----------------------------------------------------------------------------



#!/usr/bin/env python
"""
Parallel Hello World
"""

from mpi4py import MPI

size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()
name = MPI.Get_processor_name()

print(f"Hello, World! I am process {rank} of {size} on {name}.", flush=True)



import os

import mpitestutil as testutil
import mpiunittest as unittest

from mpi4py import MPI

try:
    import socket
except ImportError:
    socket = None


def ch4_ucx():
    return "ch4:ucx" in MPI.Get_library_version()


def ch4_ofi():
    return "ch4:ofi" in MPI.Get_library_version()


@unittest.skipMPI("mpich(<4.3.0)", not testutil.has_mpi_port())
@unittest.skipMPI("openmpi(<2.0.0)")
@unittest.skipMPI("openmpi(>=5.0.0,<5.0.4)")
@unittest.skipMPI("msmpi(<8.1.0)")
@unittest.skipMPI("mvapich(<3.0.0)")
@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, "mpi-world-size<2")
class TestDPM(unittest.TestCase):
    #
    message = [
        None,
        True,
        False,
        -7,
        0,
        7,
        -(2**63) + 1,
        2**63 - 1,
        -2.17,
        0.0,
        +1.62,
        1 + 2j,
        2 - 3j,
        "mpi4py",
        (1, 2, 3),
        [1, 2, 3],
        {1: 2},
    ]

    def testNamePublishing(self):
        rank = MPI.COMM_WORLD.Get_rank()
        service = f"mpi4py-{rank}"
        port = MPI.Open_port()
        MPI.Publish_name(service, port)
        found = MPI.Lookup_name(service)
        self.assertEqual(port, found)
        MPI.Unpublish_name(service, port)
        MPI.Close_port(port)

    @unittest.skipMPI("mpich(==3.4.1)", ch4_ofi())
    @unittest.skipMPI("mvapich", ch4_ofi())
    @unittest.skipMPI("impi", MPI.COMM_WORLD.Get_size() > 2)
    def testAcceptConnect(self):
        comm_self = MPI.COMM_SELF
        comm_world = MPI.COMM_WORLD
        wrank = comm_world.Get_rank()
        group_world = comm_world.Get_group()
        group = group_world.Excl([0])
        group_world.Free()
        comm = comm_world.Create(group)
        group.Free()
        if wrank == 0:
            self.assertEqual(comm, MPI.COMM_NULL)
        else:
            self.assertNotEqual(comm, MPI.COMM_NULL)
            self.assertEqual(comm.size, comm_world.size - 1)
            self.assertEqual(comm.rank, comm_world.rank - 1)
        if wrank == 0:
            port = MPI.Open_port()
            comm_world.send(port, dest=1)
            intercomm = comm_self.Accept(port)
            self.assertEqual(intercomm.remote_size, comm_world.size - 1)
            self.assertEqual(intercomm.size, 1)
            self.assertEqual(intercomm.rank, 0)
            MPI.Close_port(port)
        else:
            if wrank == 1:
                port = comm_world.recv(source=0)
            else:
                port = None
            intercomm = comm.Connect(port, root=0)
            self.assertEqual(intercomm.remote_size, 1)
            self.assertEqual(intercomm.size, comm_world.size - 1)
            self.assertEqual(intercomm.rank, comm.rank)
            comm.Free()
        if wrank == 0:
            message = TestDPM.message
            root = MPI.ROOT
        else:
            message = None
            root = 0
        message = intercomm.bcast(message, root)
        if wrank == 0:
            self.assertIsNone(message)
        else:
            self.assertEqual(message, TestDPM.message)
        intercomm.Free()

    def testConnectAccept(self):
        comm_self = MPI.COMM_SELF
        comm_world = MPI.COMM_WORLD
        wrank = comm_world.Get_rank()
        group_world = comm_world.Get_group()
        group = group_world.Excl([0])
        group_world.Free()
        comm = comm_world.Create(group)
        group.Free()
        if wrank == 0:
            self.assertEqual(comm, MPI.COMM_NULL)
        else:
            self.assertNotEqual(comm, MPI.COMM_NULL)
            self.assertEqual(comm.size, comm_world.size - 1)
            self.assertEqual(comm.rank, comm_world.rank - 1)
        if wrank == 0:
            port = comm_world.recv(source=1)
            intercomm = comm_self.Connect(port)
            self.assertEqual(intercomm.remote_size, comm_world.size - 1)
            self.assertEqual(intercomm.size, 1)
            self.assertEqual(intercomm.rank, 0)
        else:
            if wrank == 1:
                port = MPI.Open_port()
                comm_world.send(port, dest=0)
            else:
                port = None
            intercomm = comm.Accept(port, root=0)
            if wrank == 1:
                MPI.Close_port(port)
            self.assertEqual(intercomm.remote_size, 1)
            self.assertEqual(intercomm.size, comm_world.size - 1)
            self.assertEqual(intercomm.rank, comm.rank)
            comm.Free()
        if wrank == 0:
            message = TestDPM.message
            root = MPI.ROOT
        else:
            message = None
            root = 0
        message = intercomm.bcast(message, root)
        if wrank == 0:
            self.assertIsNone(message)
        else:
            self.assertEqual(message, TestDPM.message)
        intercomm.Free()

    @unittest.skipIf(socket is None, "socket")
    @unittest.skipMPI("impi", testutil.github() and os.name == "nt")
    def testJoin(self):
        rank = MPI.COMM_WORLD.Get_rank()
        server = client = address = None
        host = socket.gethostname()
        addrinfo = socket.getaddrinfo(host, None, type=socket.SOCK_STREAM)
        addr_families = [info[0] for info in addrinfo]
        # if both INET and INET6 are available, don't assume the order
        # is the same on both server and client. Prefer INET if available.
        addr_family = None
        if socket.AF_INET in addr_families:
            addr_family = socket.AF_INET
        elif socket.AF_INET6 in addr_families:
            addr_family = socket.AF_INET6
        addr_family = MPI.COMM_WORLD.bcast(addr_family, root=0)
        supported = addr_family in addr_families
        supported = MPI.COMM_WORLD.allreduce(supported, op=MPI.LAND)
        if not supported:
            self.skipTest("socket-inet")
        # create server/client sockets
        if rank == 0:  # server
            server = socket.socket(addr_family, socket.SOCK_STREAM)
            server.bind((host, 0))
            server.listen(0)
        if rank == 1:  # client
            client = socket.socket(addr_family, socket.SOCK_STREAM)
        # communicate address
        if rank == 0:
            address = server.getsockname()
            MPI.COMM_WORLD.ssend(address, 1)
        if rank == 1:
            address = MPI.COMM_WORLD.recv(None, 0)
        MPI.COMM_WORLD.Barrier()
        # stablish client/server connection
        connected = False
        if rank == 0:  # server
            client = server.accept()[0]
            server.close()
        if rank == 1:  # client
            client.connect(address)
            connected = True
        connected = MPI.COMM_WORLD.bcast(connected, root=1)
        # test Comm.Join()
        MPI.COMM_WORLD.Barrier()
        if client:
            fd = client.fileno()
            intercomm = MPI.Comm.Join(fd)
            client.close()
            if intercomm != MPI.COMM_NULL:
                self.assertEqual(intercomm.remote_size, 1)
                self.assertEqual(intercomm.size, 1)
                self.assertEqual(intercomm.rank, 0)
                if rank == 0:
                    message = TestDPM.message
                    root = MPI.ROOT
                else:
                    message = None
                    root = 0
                message = intercomm.bcast(message, root)
                if rank == 0:
                    self.assertIsNone(message)
                else:
                    self.assertEqual(message, TestDPM.message)
                intercomm.Free()
        MPI.COMM_WORLD.Barrier()


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI


def ch3_nemesis():
    return "ch3:nemesis" in MPI.Get_library_version()


@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, "mpi-world-size<2")
class BaseTestIntercomm:
    #
    BASECOMM = MPI.COMM_NULL
    INTRACOMM = MPI.COMM_NULL
    INTERCOMM = MPI.COMM_NULL

    def setUp(self):
        size = self.BASECOMM.Get_size()
        rank = self.BASECOMM.Get_rank()
        if rank < size // 2:
            self.COLOR = 0
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = size // 2
        else:
            self.COLOR = 1
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = 0
        self.INTRACOMM = self.BASECOMM.Split(self.COLOR, key=0)
        Create_intercomm = MPI.Intracomm.Create_intercomm
        self.INTERCOMM = Create_intercomm(
            self.INTRACOMM,
            self.LOCAL_LEADER,
            self.BASECOMM,
            self.REMOTE_LEADER,
        )

    def testConstructor(self):
        with self.assertRaises(TypeError):
            MPI.Intercomm(self.INTRACOMM)
        with self.assertRaises(TypeError):
            MPI.Intracomm(self.INTERCOMM)

    def tearDown(self):
        self.INTRACOMM.Free()
        self.INTERCOMM.Free()

    def testHandle(self):
        intercomm = self.INTERCOMM
        cint = intercomm.toint()
        if cint != -1:
            newcomm = MPI.Comm.fromint(cint)
            self.assertEqual(newcomm, intercomm)
            self.assertIs(type(newcomm), MPI.Intercomm)
        fint = intercomm.py2f()
        if fint != -1:
            newcomm = MPI.Comm.f2py(fint)
            self.assertEqual(newcomm, intercomm)
            self.assertIs(type(newcomm), MPI.Intercomm)

    def testLocalGroupSizeRank(self):
        intercomm = self.INTERCOMM
        local_group = intercomm.Get_group()
        self.assertEqual(local_group.size, intercomm.Get_size())
        self.assertEqual(local_group.size, intercomm.size)
        self.assertEqual(local_group.rank, intercomm.Get_rank())
        self.assertEqual(local_group.rank, intercomm.rank)
        local_group.Free()

    def testRemoteGroupSize(self):
        intercomm = self.INTERCOMM
        remote_group = intercomm.Get_remote_group()
        self.assertEqual(remote_group.size, intercomm.Get_remote_size())
        self.assertEqual(remote_group.size, intercomm.remote_size)
        remote_group.Free()

    def testMerge(self):
        basecomm = self.BASECOMM
        intercomm = self.INTERCOMM
        if basecomm.rank < basecomm.size // 2:
            high = False
        else:
            high = True
        intracomm = intercomm.Merge(high)
        self.assertEqual(intracomm.size, basecomm.size)
        self.assertEqual(intracomm.rank, basecomm.rank)
        intracomm.Free()

    def testCreateFromGroups(self):
        lgroup = self.INTERCOMM.Get_group()
        rgroup = self.INTERCOMM.Get_remote_group()
        try:
            Create_from_groups = MPI.Intercomm.Create_from_groups
            intercomm = Create_from_groups(lgroup, 0, rgroup, 0)
        except NotImplementedError:
            self.assertLess(MPI.Get_version(), (4, 0))
            self.skipTest("mpi-comm-create_from_group")
        except MPI.Exception as exc:
            UNSUPPORTED = MPI.ERR_UNSUPPORTED_OPERATION
            if exc.Get_error_class() != UNSUPPORTED:
                raise
        else:
            ccmp = MPI.Comm.Compare(self.INTERCOMM, intercomm)
            intercomm.Free()
            self.assertEqual(ccmp, MPI.CONGRUENT)
        finally:
            lgroup.Free()
            rgroup.Free()

    def testSplit(self):
        base = self.INTERCOMM
        comm = base.Split(42, 42)
        self.assertEqual(comm.Get_rank(), base.Get_rank())
        self.assertEqual(comm.Get_size(), base.Get_size())
        self.assertEqual(comm.Get_remote_size(), base.Get_remote_size())
        comm.Free()
        color = base.Get_rank()
        comm = base.Split(color, 42)
        if comm != MPI.COMM_NULL:
            self.assertEqual(comm.Get_rank(), 0)
            self.assertEqual(comm.Get_size(), 1)
            self.assertEqual(comm.Get_remote_size(), 1)
            comm.Free()

    @unittest.skipMPI("msmpi")
    @unittest.skipMPI("openmpi(<5.0.0)")
    @unittest.skipMPI("mpich(<4.2.0)", ch3_nemesis())
    @unittest.skipMPI("mvapich(<3.0.0)", MPI.COMM_WORLD.Get_size() > 2)
    def testSplitTypeShared(self):
        try:
            comm = self.INTERCOMM.Split_type(MPI.COMM_TYPE_SHARED)
        except NotImplementedError:
            self.skipTest("mpi-comm-split_type")
        if comm != MPI.COMM_NULL:
            comm.Free()
        comm = self.INTERCOMM.Split_type(MPI.UNDEFINED)
        self.assertEqual(comm, MPI.COMM_NULL)

    def testPyProps(self):
        comm = self.INTERCOMM
        #
        self.assertEqual(comm.rank, comm.Get_rank())
        self.assertEqual(comm.size, comm.Get_size())
        self.assertEqual(comm.remote_size, comm.Get_remote_size())
        #
        group = comm.remote_group
        self.assertEqual(type(group), MPI.Group)
        group.Free()
        #
        info = comm.info
        self.assertEqual(type(info), MPI.Info)
        info.Free()
        #
        self.assertTrue(comm.is_inter)
        self.assertFalse(comm.is_intra)


class TestIntercomm(BaseTestIntercomm, unittest.TestCase):
    #
    BASECOMM = MPI.COMM_WORLD


class TestIntercommDup(TestIntercomm):
    #
    def setUp(self):
        self.BASECOMM = self.BASECOMM.Dup()
        super().setUp()

    def tearDown(self):
        self.BASECOMM.Free()
        super().tearDown()


class TestIntercommDupDup(TestIntercomm):
    #
    def setUp(self):
        super().setUp()
        INTERCOMM = self.INTERCOMM
        self.INTERCOMM = self.INTERCOMM.Dup()
        INTERCOMM.Free()


@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, "mpi-world-size<2")
class TestIntercommCreateFromGroups(unittest.TestCase):
    #
    @unittest.skipMPI("openmpi", MPI.COMM_WORLD.Get_size() > 2)
    def testPair(self):
        done = True
        rank = MPI.COMM_WORLD.Get_rank()
        if rank < 2:
            world_group = MPI.COMM_WORLD.Get_group()
            local_group = world_group.Incl([rank])
            remote_group = world_group.Incl([1 - rank])
            world_group.Free()
            try:
                comm = MPI.Intercomm.Create_from_groups(
                    local_group,
                    0,
                    remote_group,
                    0,
                )
                self.assertEqual(comm.Get_size(), 1)
                self.assertEqual(comm.Get_remote_size(), 1)
                comm.Free()
            except NotImplementedError:
                done = False
            finally:
                local_group.Free()
                remote_group.Free()
        done = MPI.COMM_WORLD.allreduce(done, op=MPI.LAND)
        if not done:
            self.assertLess(MPI.Get_version(), (4, 0))
            self.skipTest("mpi-intercomm-create_from_groups")

    def testHalf(self):
        done = True
        size = MPI.COMM_WORLD.Get_size()
        rank = MPI.COMM_WORLD.Get_rank()
        world_group = MPI.COMM_WORLD.Get_group()
        low_group = world_group.Range_incl([(0, size // 2 - 1, 1)])
        high_group = world_group.Range_incl([(size // 2, size - 1, 1)])
        world_group.Free()
        if rank <= size // 2 - 1:
            local_group, remote_group = low_group, high_group
            local_leader, remote_leader = 0, high_group.Get_size() - 1
        else:
            local_group, remote_group = high_group, low_group
            local_leader, remote_leader = high_group.Get_size() - 1, 0
        try:
            comm = MPI.Intercomm.Create_from_groups(
                local_group,
                local_leader,
                remote_group,
                remote_leader,
            )
            self.assertEqual(comm.Get_rank(), local_group.Get_rank())
            self.assertEqual(comm.Get_size(), local_group.Get_size())
            self.assertEqual(comm.Get_remote_size(), remote_group.Get_size())
            comm.Free()
        except NotImplementedError:
            done = False
        finally:
            local_group.Free()
            remote_group.Free()
        if not done:
            self.assertLess(MPI.Get_version(), (4, 0))
            self.skipTest("mpi-intercomm-create_from_groups")


if __name__ == "__main__":
    unittest.main()
